from django.core.management import BaseCommand
from meta_pharmacy.utils import update_initial_pharmacy_data


class Command(BaseCommand):

    def handle(self, *args, **options):
        """This is an example of an initial set of data"""
        update_initial_pharmacy_data()
        print("Done")



from django.conf import settings

from .dashboard_templates import dashboard_templates


class DashboardMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_view(self, request, *args) -> None:
        template_data = dashboard_templates
        try:
            template_data.update(settings.REVIEW_DASHBOARD_BASE_TEMPLATES)
        except AttributeError:
            pass
        request.template_data.update(**template_data)

    def process_template_response(self, request, response):
        if response.context_data:
            response.context_data.update(**request.template_data)
        return response



from ..ordered_collection import OrderedCollection
from ..schedule import Schedule


class SchedulesCollectionError(Exception):
    pass


class SchedulesCollection(OrderedCollection):
    key = "name"
    ordering_attr = "sequence"

    def __init__(self, visit_schedule_name: str, *args, **kwargs) -> None:
        self.visit_schedule_name = visit_schedule_name
        super().__init__(*args, **kwargs)

    def get_schedule(
        self, model: str | None = None, schedule_name: str | None = None
    ) -> Schedule:
        """Returns a schedule or raises; by name, by onschedule/offschedule model
        or by model label_lower.
        """
        schedule = None
        if model:
            model = model.lower()
            for item in self.values():
                if model in (item.onschedule_model, item.offschedule_model):
                    schedule = item
                if schedule:
                    break
        elif schedule_name:
            schedule = self.get(schedule_name)
        if not schedule:
            raise SchedulesCollectionError(
                f"Schedule does not exist. Using model={model}, schedule_name={schedule_name}."
            )
        return schedule

    def validate(self) -> None:
        for schedule in self.values():
            schedule.validate(visit_schedule_name=self.visit_schedule_name)



from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

from .auth_objects import export_codenames
from .constants import DATA_EXPORTER_ROLE, EXPORT, EXPORT_PII


def update_site_auths():
    site_auths.add_post_update_func(
        "edc_export",
        remove_default_model_permissions_from_edc_permissions,
    )

    site_auths.add_custom_permissions_tuples(
        model="edc_export.edcpermissions",
        codename_tuples=[
            (
                "edc_export.export_visitschedule",
                "Can export the visit schedule",
            ),
            (
                "edc_export.view_export_dashboard",
                "Can view export dashboard",
            ),
            (
                "edc_export.export_subjectschedulehistory",
                "Can export subject schedule history",
            ),
            (
                "edc_export.export_pii",
                "Can export PII",
            ),
        ],
    )

    site_auths.add_group(*export_codenames, name=EXPORT)
    site_auths.add_role(EXPORT, name=DATA_EXPORTER_ROLE)
    site_auths.add_group("edc_export.export_pii", name=EXPORT_PII)


update_site_auths()



from __future__ import annotations

import csv
import sys
from pathlib import Path
from pprint import pprint
from uuid import uuid4

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.color import color_style
from tqdm import tqdm

from edc_sites.site import sites as site_sites

from .randomization_list_verifier import RandomizationListVerifier

style = color_style()


class RandomizationListImportError(Exception):
    pass


class RandomizationListAlreadyImported(Exception):
    pass


class InvalidAssignment(Exception):
    pass


class RandomizationListImporter:
    """Imports upon instantiation a formatted randomization CSV file
    into model RandomizationList.

    default CSV file is the project's randomization_list.csv

    name: name of randomizer, e.g. "default"

    To import SIDS from CSV for the first time:

        from edc_randomization.randomization_list_importer import RandomizationListImporter

        RandomizationListImporter(name='default', add=False, dryrun=False)

        # note: if this is not the first time you will get:
        # RandomizationListImportError: Not importing CSV.
        # edc_randomization.randomizationlist model is not empty!

    To add additional sids from CSV without touching existing model instances:

        from edc_randomization.randomization_list_importer import RandomizationListImporter

        RandomizationListImporter(name='default', add=True, dryrun=False)


    Format:
        sid,site_name, assignment, description, orig_site, orig_allocation, orig_desc
        1,gaborone,intervention,single_dose
        2,gaborone,control,two_doses
        ...
    """

    required_csv_fieldnames = ("sid", "assignment", "site_name", "description")
    verifier_cls = RandomizationListVerifier

    def __init__(
        self,
        randomizer_model_cls=None,
        randomizer_name: str = None,
        randomizationlist_path: Path | str = None,
        assignment_map: dict[str, int] = None,
        verbose: bool = None,
        overwrite: bool = None,
        add: bool = None,
        dryrun: bool = None,
        username: str = None,
        revision: str = None,
        sid_count_for_tests: int = None,
        extra_csv_fieldnames: list[str] | None = None,
        **kwargs,
    ):
        self.verify_messages: str | None = None
        self.add = add
        self.overwrite = overwrite
        self.verbose = True if verbose is None else verbose
        self.dryrun = dryrun
        self.revision = revision
        self.user = username
        self.sid_count_for_tests = sid_count_for_tests
        self.randomizer_model_cls = randomizer_model_cls
        self.randomizer_name = randomizer_name
        self.assignment_map = assignment_map
        self.randomizationlist_path: Path = Path(randomizationlist_path).expanduser()
        self.required_csv_fieldnames.extend(extra_csv_fieldnames or [])

        if self.dryrun:
            sys.stdout.write(
                style.MIGRATE_HEADING("\n ->> Dry run. No changes will be made.\n")
            )
        if self.verbose and add:
            count = self.randomizer_model_cls.objects.all().count()
            sys.stdout.write(
                style.SUCCESS(
                    f"\n(*) Randolist model has {count} SIDs (count before import).\n"
                )
            )

    def import_list(self, **kwargs) -> tuple[int, Path]:
        """Imports CSV and verifies."""
        self._raise_on_empty_file()
        self._raise_on_invalid_header()
        self._raise_on_already_imported()
        self._raise_on_duplicates()
        if self.verbose:
            sys.stdout.write(
                style.SUCCESS(
                    "\nImport CSV data\n"
                    "  Randomizer:\n"
                    f"    -  Name: {self.randomizer_name}\n"
                    f"    -  Assignments: {self.assignment_map}\n"
                    f"    -  Model: {self.randomizer_model_cls._meta.label_lower}\n"
                    f"    -  Path: {self.randomizationlist_path}\n"
                )
            )
        rec_count = self._import_csv_to_model()
        self.verify_messages = self._verify_data(**kwargs)
        self._summarize_results()
        if self.verbose:
            sys.stdout.write(
                style.SUCCESS("\nDone.------------------------------------------------\n")
            )
        return rec_count, self.randomizationlist_path

    def _summarize_results(self):
        if self.verbose:
            count = self.randomizer_model_cls.objects.all().count()
            msg = (
                f"\n    - Imported {count} SIDs for randomizer "
                f"`{self.randomizer_name}` into model "
                f"`{self.randomizer_model_cls._meta.label_lower}` \n"
                f"      from {self.randomizationlist_path}.\n"
            )
            sys.stdout.write(style.SUCCESS(msg))
            if self.verify_messages:
                sys.stdout.write(style.ERROR("\n    ! Verification failed. "))
            else:
                sys.stdout.write(style.SUCCESS("    - Verified OK. \n"))

    def _raise_on_empty_file(self):
        if self.randomizationlist_path.stat().st_size < 1:
            raise RandomizationListImportError(
                f"File is empty. See {self.randomizer_name}. "
                f"Got {self.randomizationlist_path} (1)."
            )
        index = 0
        with self.randomizationlist_path.open(mode="r") as csvfile:
            reader = csv.DictReader(csvfile)
            for index, row in enumerate(reader):
                if index == 0:
                    continue
        if index == 0:
            raise RandomizationListImportError(
                f"File is empty. See {self.randomizer_name}. "
                f"Got {self.randomizationlist_path} (2)."
            )

    def _raise_on_invalid_header(self):
        with self.randomizationlist_path.open(mode="r") as csvfile:
            reader = csv.DictReader(csvfile)
            for index, row in enumerate(reader):
                if index == 0:
                    for fieldname in self.required_csv_fieldnames:
                        if fieldname not in row:
                            raise RandomizationListImportError(
                                f"Invalid header. Missing column `{fieldname}`. Got {row}"
                            )
                elif index == 1:
                    if self.dryrun:
                        row_as_dict = {k: v for k, v in row.items()}
                        print(" -->  First row:")
                        print(f" -->  {list(row_as_dict.keys())}")
                        print(f" -->  {list(row_as_dict.values())}")
                        obj = self.randomizer_model_cls(**self.get_import_options(row))
                        pprint(obj.__dict__)
                else:
                    break

    def _raise_on_already_imported(self):
        if not self.dryrun:
            if self.overwrite:
                self.randomizer_model_cls.objects.all().delete()
            if self.randomizer_model_cls.objects.all().count() > 0 and not self.add:
                raise RandomizationListAlreadyImported(
                    "Not importing CSV. "
                    f"{self.randomizer_model_cls._meta.label_lower} "
                    "model is not empty!"
                )

    def get_sid_list(self) -> list[int]:
        with self.randomizationlist_path.open(mode="r") as csvfile:
            reader = csv.DictReader(csvfile)
            sids = [int(row["sid"]) for row in reader]
        if len(sids) != len(list(set(sids))):
            raise RandomizationListImportError("Invalid file. Detected duplicate SIDs")
        return sids

    def _raise_on_duplicates(self) -> None:
        self.get_sid_list()

    def _import_csv_to_model(self) -> int:
        """Imports a CSV to populate the "rando" model"""
        objs = []
        rec_count = 0

        if self.sid_count_for_tests is not None:
            sys.stdout.write(
                style.WARNING(
                    "\nNote: Importing a `subset` of the randomization list for tests "
                    f"({self.sid_count_for_tests}).\n"
                )
            )
            sid_count = self.sid_count_for_tests
        else:
            sid_count = len(self.get_sid_list())
        with self.randomizationlist_path.open(mode="r") as csvfile:
            reader = csv.DictReader(csvfile)
            for row in tqdm(reader, total=sid_count):
                if self.sid_count_for_tests and len(objs) == self.sid_count_for_tests:
                    break
                row = {k: v.strip() for k, v in row.items()}
                try:
                    self.randomizer_model_cls.objects.get(sid=row["sid"])
                except ObjectDoesNotExist:
                    opts = self.get_import_options(row)
                    opts.update(self.get_extra_import_options(row))
                    if self.user:
                        opts.update(user_created=self.user)
                    if self.revision:
                        opts.update(revision=self.revision)
                    obj = self.randomizer_model_cls(**opts)
                    objs.append(obj)
            if not self.dryrun:
                self.randomizer_model_cls.objects.bulk_create(objs)
                rec_count = self.randomizer_model_cls.objects.all().count()
                if not sid_count == rec_count:
                    raise RandomizationListImportError(
                        "Incorrect record count on import. "
                        f"Expected {sid_count}. Got {rec_count}."
                    )
            else:
                sys.stdout.write(
                    style.MIGRATE_HEADING(
                        "\n ->> this is a dry run. No changes were saved. **\n"
                    )
                )
        return rec_count

    def _verify_data(self, **kwargs) -> list[str]:
        verifier = self.verifier_cls(
            assignment_map=self.assignment_map,
            randomizationlist_path=self.randomizationlist_path,
            randomizer_model_cls=self.randomizer_model_cls,
            randomizer_name=self.randomizer_name,
            required_csv_fieldnames=self.required_csv_fieldnames,
            **kwargs,
        )
        return verifier.messages

    def get_assignment(self, row: dict, assignment_map: dict[str, int]) -> str:
        """Returns assignment (text) after checking validity."""
        return self.valid_assignment_or_raise(row["assignment"], assignment_map)

    def get_allocation(self, row: dict, assignment_map: dict[str, int]) -> int:
        """Returns an integer allocation for the given
        assignment or raises.
        """
        assignment = self.get_assignment(row, assignment_map)
        return assignment_map.get(assignment)

    def valid_assignment_or_raise(
        self, assignment: str, assignment_map: dict[str, int]
    ) -> str:
        if assignment not in assignment_map:
            raise InvalidAssignment(
                f"Invalid assignment. Expected one of {list(assignment_map.keys())}. "
                f"Got `{assignment}`. "
                f"See randomizer `{self.randomizer_name}` {self!r}. "
            )
        return assignment

    def get_import_options(self, row):
        return dict(
            id=uuid4(),
            sid=row["sid"],
            assignment=self.get_assignment(row, self.assignment_map),
            allocation=str(self.get_allocation(row, self.assignment_map)),
            randomizer_name=self.randomizer_name,
            site_name=self.validate_site_name(row),
            **self.get_extra_import_options(row),
        )

    def get_extra_import_options(self, row):
        return {}

    @staticmethod
    def get_site_names() -> dict[str, str]:
        """A dict of site names for the target randomizer."""
        return {
            single_site.name: single_site.name for single_site in site_sites.all().values()
        }

    def validate_site_name(self, row) -> str:
        """Returns the site name or raises"""
        try:
            site_name = self.get_site_names()[row["site_name"].lower()]
        except KeyError:
            raise RandomizationListImportError(
                f"Invalid site. Got {row['site_name']}. "
                f"Expected one of {self.get_site_names().keys()}"
            )
        return site_name



"""To customize any of the values below,
use settings.LAB_DASHBOARD_BASE_TEMPLATES.
"""

from django.conf import settings

dashboard_templates = dict(
    aliquot_listboard_template="edc_lab_dashboard/aliquot_listboard.html",
    box_listboard_template="edc_lab_dashboard/box_listboard.html",
    edc_lab_base_template="edc_lab_dashboard/base.html",
    home_template="edc_lab_dashboard/home.html",
    edc_lab_base_listboard_template="edc_lab_dashboard/base_listboard.html",
    edc_lab_listboard_template="edc_lab_dashboard/listboard.html",
    manage_box_listboard_template="edc_lab_dashboard/manage_box_listboard.html",
    manage_manifest_listboard_template="edc_lab_dashboard/manage_manifest_listboard.html",
    manifest_listboard_template="edc_lab_dashboard/manifest_listboard.html",
    pack_listboard_template="edc_lab_dashboard/pack_listboard.html",
    process_listboard_template="edc_lab_dashboard/process_listboard.html",
    receive_listboard_template="edc_lab_dashboard/receive_listboard.html",
    requisition_listboard_template="edc_lab_dashboard/requisition_listboard.html",
    result_listboard_template="edc_lab_dashboard/result_listboard.html",
    verify_box_listboard_template="edc_lab_dashboard/verify_box_listboard.html",
)

try:
    dashboard_templates.update(**settings.LAB_DASHBOARD_BASE_TEMPLATES)
except AttributeError:
    pass



# Generated by Django 5.2.3 on 2025-07-08 12:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0084_confirmationatsiteitem_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="confirmationatsite",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Stock Confirmation at Site ",
                "verbose_name_plural": "Stock Confirmations at Site",
            },
        ),
        migrations.AlterModelOptions(
            name="confirmationatsiteitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Stock Confirmation at Site Item",
                "verbose_name_plural": "Stock Confirmation at Site Items",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconfirmationatsite",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Stock Confirmation at Site ",
                "verbose_name_plural": "historical Stock Confirmations at Site",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconfirmationatsiteitem",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Stock Confirmation at Site Item",
                "verbose_name_plural": "historical Stock Confirmation at Site Items",
            },
        ),
        migrations.RenameField(
            model_name="confirmationatsiteitem",
            old_name="stock_transfer_confirmation",
            new_name="confirmation_at_site",
        ),
        migrations.RenameField(
            model_name="historicalconfirmationatsiteitem",
            old_name="stock_transfer_confirmation",
            new_name="confirmation_at_site",
        ),
        migrations.RenameIndex(
            model_name="confirmationatsite",
            new_name="edc_pharmac_modifie_0b23c2_idx",
            old_name="edc_pharmac_modifie_88e9fc_idx",
        ),
        migrations.RenameIndex(
            model_name="confirmationatsite",
            new_name="edc_pharmac_user_mo_2f1be6_idx",
            old_name="edc_pharmac_user_mo_9c1939_idx",
        ),
        migrations.RenameIndex(
            model_name="confirmationatsiteitem",
            new_name="edc_pharmac_modifie_d7c6b9_idx",
            old_name="edc_pharmac_modifie_84ddbd_idx",
        ),
        migrations.RenameIndex(
            model_name="confirmationatsiteitem",
            new_name="edc_pharmac_user_mo_0acc89_idx",
            old_name="edc_pharmac_user_mo_956b50_idx",
        ),
        migrations.RemoveField(
            model_name="historicalstock",
            name="confirmed_at_site",
        ),
        migrations.RemoveField(
            model_name="historicalstock",
            name="dispense",
        ),
        migrations.RemoveField(
            model_name="historicalstock",
            name="transferred",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="confirmed_at_site",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="dispense",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="transferred",
        ),
        migrations.AlterField(
            model_name="historicalstoragebinitem",
            name="stock",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                limit_choices_to={
                    "confirmationatsiteitem__isnull": False,
                    "dispenseitem__isnull": True,
                },
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.stock",
            ),
        ),
        migrations.AlterField(
            model_name="storagebinitem",
            name="stock",
            field=models.ForeignKey(
                limit_choices_to={
                    "confirmationatsiteitem__isnull": False,
                    "dispenseitem__isnull": True,
                },
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.stock",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-11-26 11:56
from __future__ import unicode_literals

import django.core.validators
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_extensions.db.fields
import django_revision.revision_field
from django.conf import settings
from django.db import migrations, models

import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("edc_appointment", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Appointment",
            fields=[
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        default="mac2-2.local",
                        editable=False,
                        help_text="System field. (modified on create only)",
                        max_length=50,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        editable=False,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(
                        editable=False, max_length=50, verbose_name="Subject Identifier"
                    ),
                ),
                (
                    "timepoint_status",
                    models.CharField(
                        choices=[
                            ("open", "Open"),
                            ("feedback", "Feedback"),
                            ("closed", "Closed"),
                        ],
                        default="open",
                        max_length=15,
                    ),
                ),
                (
                    "timepoint_opened_datetime",
                    models.DateTimeField(
                        editable=False,
                        help_text="the original calculated model's datetime, updated in the signal",
                        null=True,
                    ),
                ),
                (
                    "timepoint_closed_datetime",
                    models.DateTimeField(editable=False, null=True),
                ),
                (
                    "visit_schedule_name",
                    models.CharField(
                        editable=False,
                        help_text='the name of the visit schedule used to find the "schedule"',
                        max_length=25,
                    ),
                ),
                ("schedule_name", models.CharField(editable=False, max_length=25)),
                (
                    "visit_code",
                    models.CharField(editable=False, max_length=25, null=True),
                ),
                (
                    "timepoint",
                    models.DecimalField(
                        decimal_places=1,
                        editable=False,
                        help_text="timepoint from schedule",
                        max_digits=6,
                        null=True,
                    ),
                ),
                (
                    "timepoint_datetime",
                    models.DateTimeField(
                        editable=False,
                        help_text="Unadjusted datetime calculated from visit schedule",
                        null=True,
                    ),
                ),
                (
                    "appt_close_datetime",
                    models.DateTimeField(
                        editable=False,
                        help_text="timepoint_datetime adjusted according to the nearest available datetime for this facility",
                        null=True,
                    ),
                ),
                ("facility_name", models.CharField(max_length=25)),
                (
                    "visit_instance",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        default="0",
                        help_text="A decimal to represent an additional report to be included with the original visit report. (NNNN.0)",
                        max_length=1,
                        null=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                "[0-9]", "Must be a number from 0-9"
                            )
                        ],
                        verbose_name="Instance",
                    ),
                ),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        blank=True,
                        default=0,
                        help_text="An integer to represent the sequence of additional appointments relative to the base appointment, 0, needed to complete data collection for the timepoint. (NNNN.0)",
                        null=True,
                        verbose_name="Sequence",
                    ),
                ),
                (
                    "appt_datetime",
                    models.DateTimeField(
                        db_index=True, verbose_name="Appointment date and time"
                    ),
                ),
                (
                    "appt_type",
                    models.CharField(
                        choices=[
                            ("clinic", "In clinic"),
                            ("telephone", "By telephone"),
                            ("home", "At home"),
                        ],
                        default="clinic",
                        help_text="Default for subject may be edited Subject Configuration.",
                        max_length=20,
                        verbose_name="Appointment type",
                    ),
                ),
                (
                    "appt_status",
                    models.CharField(
                        choices=[
                            ("new", "New"),
                            ("in_progress", "In Progress"),
                            ("incomplete", "Incomplete"),
                            ("done", "Done"),
                            ("cancelled", "Cancelled"),
                        ],
                        db_index=True,
                        default="new",
                        max_length=25,
                        verbose_name="Status",
                    ),
                ),
                (
                    "appt_reason",
                    models.CharField(
                        blank=True,
                        help_text="Reason for appointment",
                        max_length=25,
                        verbose_name="Reason for appointment",
                    ),
                ),
                (
                    "comment",
                    models.CharField(blank=True, max_length=250, verbose_name="Comment"),
                ),
                ("is_confirmed", models.BooleanField(default=False, editable=False)),
                (
                    "consent_version",
                    models.CharField(default="?", editable=False, max_length=10),
                ),
            ],
            options={"abstract": False, "ordering": ("appt_datetime",)},
        ),
        migrations.CreateModel(
            name="HistoricalAppointment",
            fields=[
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        default="mac2-2.local",
                        editable=False,
                        help_text="System field. (modified on create only)",
                        max_length=50,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        editable=False,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(
                        editable=False, max_length=50, verbose_name="Subject Identifier"
                    ),
                ),
                (
                    "timepoint_status",
                    models.CharField(
                        choices=[
                            ("open", "Open"),
                            ("feedback", "Feedback"),
                            ("closed", "Closed"),
                        ],
                        default="open",
                        max_length=15,
                    ),
                ),
                (
                    "timepoint_opened_datetime",
                    models.DateTimeField(
                        editable=False,
                        help_text="the original calculated model's datetime, updated in the signal",
                        null=True,
                    ),
                ),
                (
                    "timepoint_closed_datetime",
                    models.DateTimeField(editable=False, null=True),
                ),
                (
                    "visit_schedule_name",
                    models.CharField(
                        editable=False,
                        help_text='the name of the visit schedule used to find the "schedule"',
                        max_length=25,
                    ),
                ),
                ("schedule_name", models.CharField(editable=False, max_length=25)),
                (
                    "visit_code",
                    models.CharField(editable=False, max_length=25, null=True),
                ),
                (
                    "timepoint",
                    models.DecimalField(
                        decimal_places=1,
                        editable=False,
                        help_text="timepoint from schedule",
                        max_digits=6,
                        null=True,
                    ),
                ),
                (
                    "timepoint_datetime",
                    models.DateTimeField(
                        editable=False,
                        help_text="Unadjusted datetime calculated from visit schedule",
                        null=True,
                    ),
                ),
                (
                    "appt_close_datetime",
                    models.DateTimeField(
                        editable=False,
                        help_text="timepoint_datetime adjusted according to the nearest available datetime for this facility",
                        null=True,
                    ),
                ),
                ("facility_name", models.CharField(max_length=25)),
                (
                    "visit_instance",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        default="0",
                        help_text="A decimal to represent an additional report to be included with the original visit report. (NNNN.0)",
                        max_length=1,
                        null=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                "[0-9]", "Must be a number from 0-9"
                            )
                        ],
                        verbose_name="Instance",
                    ),
                ),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        blank=True,
                        default=0,
                        help_text="An integer to represent the sequence of additional appointments relative to the base appointment, 0, needed to complete data collection for the timepoint. (NNNN.0)",
                        null=True,
                        verbose_name="Sequence",
                    ),
                ),
                (
                    "appt_datetime",
                    models.DateTimeField(
                        db_index=True, verbose_name="Appointment date and time"
                    ),
                ),
                (
                    "appt_type",
                    models.CharField(
                        choices=[
                            ("clinic", "In clinic"),
                            ("telephone", "By telephone"),
                            ("home", "At home"),
                        ],
                        default="clinic",
                        help_text="Default for subject may be edited Subject Configuration.",
                        max_length=20,
                        verbose_name="Appointment type",
                    ),
                ),
                (
                    "appt_status",
                    models.CharField(
                        choices=[
                            ("new", "New"),
                            ("in_progress", "In Progress"),
                            ("incomplete", "Incomplete"),
                            ("done", "Done"),
                            ("cancelled", "Cancelled"),
                        ],
                        db_index=True,
                        default="new",
                        max_length=25,
                        verbose_name="Status",
                    ),
                ),
                (
                    "appt_reason",
                    models.CharField(
                        blank=True,
                        help_text="Reason for appointment",
                        max_length=25,
                        verbose_name="Reason for appointment",
                    ),
                ),
                (
                    "comment",
                    models.CharField(blank=True, max_length=250, verbose_name="Comment"),
                ),
                ("is_confirmed", models.BooleanField(default=False, editable=False)),
                (
                    "consent_version",
                    models.CharField(default="?", editable=False, max_length=10),
                ),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        primary_key=True, serialize=False
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "get_latest_by": "history_date",
                "verbose_name": "historical appointment",
                "ordering": ("-history_date", "-history_id"),
            },
        ),
        migrations.AlterUniqueTogether(
            name="appointment",
            unique_together=set(
                [
                    (
                        "subject_identifier",
                        "visit_schedule_name",
                        "schedule_name",
                        "visit_code",
                        "visit_code_sequence",
                    ),
                    (
                        "subject_identifier",
                        "visit_schedule_name",
                        "schedule_name",
                        "visit_code",
                        "timepoint",
                    ),
                ]
            ),
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-22 01:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_auth", "0009_userprofile_alternate_email")]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="alternate_email",
            field=models.EmailField(
                blank=True,
                max_length=254,
                null=True,
                verbose_name="Alternate email address",
            ),
        )
    ]



from django.db import models
from django.db.models import Index

from edc_model.models import BaseUuidModel, HistoricalRecords


class ExportReceiptManager(models.Manager):
    def get_by_natural_key(self, export_uuid):
        return self.get(export_uuid=export_uuid)


class ExportReceipt(BaseUuidModel):
    export_uuid = models.UUIDField(
        editable=False, help_text="system field for export tracking."
    )

    model = models.CharField(max_length=64)

    tx_pk = models.CharField(max_length=36)

    timestamp = models.CharField(max_length=50, null=True)

    received_datetime = models.DateTimeField(null=True, help_text="date ACK received")

    objects = ExportReceiptManager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.model_name}.{self.export_uuid}"

    def natural_key(self):
        return (self.export_uuid,)

    class Meta(BaseUuidModel.Meta):
        indexes = (Index(fields=["timestamp"]),)



from django.contrib import admin
from django.utils.html import format_html
from django.utils.safestring import mark_safe


class HealthFacilityModelAdminMixin:
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "report_datetime",
                    "name",
                    "health_facility_type",
                    "health_facility_type_other",
                )
            },
        ),
        (
            "Clinic days",
            {"fields": ("mon", "tue", "wed", "thu", "fri", "sat")},
        ),
        (
            "Location",
            {
                "description": "Provide this information if available",
                "fields": (
                    "gps",
                    "latitude",
                    "longitude",
                ),
            },
        ),
        (
            "Notes",
            {
                "fields": ("notes",),
            },
        ),
    )

    list_display = (
        "facility_name",
        "health_facility_type",
        "clinic_days",
        "map",
    )

    list_filter = (
        "report_datetime",
        "health_facility_type",
    )

    radio_fields = {
        "health_facility_type": admin.VERTICAL,
    }

    search_fields = (
        "name",
        "health_facility_type__name",
    )

    @admin.display(description="Name", ordering="name")
    def facility_name(self, obj=None):
        return format_html(
            '<span style="white-space:nowrap;">{name}</span>',
            name=obj.name,
        )

    @admin.display(description="Map")
    def map(self, obj=None):
        if obj.latitude and obj.longitude:
            return format_html(
                '<A href="{url}@{latitude},{longitude},15z">'
                '<i class="fas fa-location-dot"></i>',
                url=mark_safe("https://www.google.com/maps/"),  # nosec B703, B308
                latitude=obj.latitude,
                longitude=obj.longitude,
            )
        return None

    @admin.display(description="Days")
    def clinic_days(self, obj=None) -> str:
        return format_html(
            '<span style="white-space:nowrap;">{clinic_days_str}</span>',
            clinic_days_str=obj.clinic_days_str,
        )



from __future__ import annotations

from datetime import datetime

from edc_utils import to_utc


class ReportDatetimeModelFormMixin:
    # may also be appt_datetime or requisition_datetime
    report_datetime_field_attr: str = "report_datetime"

    @property
    def report_datetime(self) -> datetime | None:
        """Returns the report_datetime in UTC or None from
        cleaned_data.

        if key does not exist, returns the instance report_datetime.

        Django should raise a required field ValidationError before getting here
        if report_datetime is none.
        """
        report_datetime = None
        if self.report_datetime_field_attr in self.cleaned_data:
            report_datetime = self.cleaned_data.get(self.report_datetime_field_attr)
            if report_datetime:
                report_datetime = to_utc(report_datetime)
        elif self.instance:
            report_datetime = getattr(self.instance, self.report_datetime_field_attr)
        return report_datetime



from edc_list_data.model_mixins import ListModelMixin
from edc_model.models import BaseUuidModel
from edc_sites.managers import CurrentSiteManager

from .model_mixins import SubjectTransferModelMixin


class TransferReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Transfer Reasons"
        verbose_name_plural = "Transfer Reasons"


class SubjectTransfer(SubjectTransferModelMixin, BaseUuidModel):

    on_site = CurrentSiteManager()

    class Meta(SubjectTransferModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Subject transfer"
        verbose_name_plural = "Subject transfers"



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

UserModel = get_user_model()


class ModelBackendWithSite(ModelBackend):
    """An authentication backend to only allow a login
    associated with the current SITE_ID.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        user = super().authenticate(request, username=username, password=password)
        if user:
            sites = [obj.id for obj in user.userprofile.sites.all()]
            try:
                site_id = request.site.id
            except AttributeError:
                site_id = settings.SITE_ID
            if user.is_superuser or site_id in sites:
                return user
        return None



from __future__ import annotations

from datetime import date, datetime

from django.conf import settings
from django.core.exceptions import ValidationError

from edc_constants.constants import GT, GTE, LT, LTE
from edc_model import estimated_date_from_ago
from edc_utils import convert_php_dateformat

FAILED_COMPARISON = "FAILED_COMPARISON"
BEFORE_AFTER_BOTH_TRUE = "BEFORE_AFTER_BOTH_TRUE"
BEFORE_AFTER_BOTH_FALSE = "BEFORE_AFTER_BOTH_FALSE"
MISSING_REPORT_DATETIME = "MISSING_REPORT_DATETIME"
MISSING_DATE_AND_AGO = "MISSING_DATE_AND_AGO"
DATE_AND_AGO_CONFLICT = "DATE_AND_AGO_CONFLICT"


class MedicalDateError(ValidationError):
    def __init__(self, message: dict, code: str = None, params=None):
        self.code = code
        super().__init__(message, code, params)

    @property
    def message_dict(self):
        self.error_dict
        return {k: v[0] for k, v in dict(self).items()}


class MedicalDate(date):
    _after_reference: bool = None
    _before_reference: bool = None
    _cleaned_data: dict = None
    _date_field: str = None
    _ago_field: str = None
    _field: str = None
    _inclusive: bool = None
    _label: str = None
    _op: str = None
    _reference_date: date | datetime | MedicalDate = None
    _value: date = None
    _word: str = None

    def __new__(
        cls,
        date_field: str,
        ago_field: str,
        cleaned_data: dict,
        reference_date: date | datetime | MedicalDate,
        reference_is_none_msg: str,
        before_reference: bool | None = None,
        after_reference: bool | None = None,
        inclusive: bool | None = None,
        label: str | None = None,
    ):
        cls._after_reference = after_reference
        cls._before_reference = before_reference
        cls._cleaned_data = cleaned_data
        cls._date_field = date_field
        cls._ago_field = ago_field
        cls._inclusive = inclusive
        cls._label = label

        cls._report_date_or_raise()
        cls._one_date_or_raise()
        cls._date_or_raise()
        try:
            cls._value = cls._cleaned_data.get(cls._date_field).date()
        except AttributeError:
            cls._value = cls._cleaned_data.get(cls._date_field)
        if cls._value:
            cls._field = cls._date_field
        else:
            cls._field = cls._ago_field
            cls._value = estimated_date_from_ago(
                cleaned_data=cls._cleaned_data, ago_field=cls._ago_field
            )
        try:
            cls._reference_date = reference_date.date()
        except AttributeError:
            cls._reference_date = reference_date
        cls._op, cls._word = cls._get_operator()
        if not cls._compare_date_and_reference():
            cls._raise_on_failed_comparison()
        if (
            cls._cleaned_data.get("report_datetime").date() != cls._reference_date
            and cls._value > cls._cleaned_data.get("report_datetime").date()
        ):
            raise MedicalDateError(
                {cls._field: "Cannot be after report date"},
                code=FAILED_COMPARISON,
            )
        return super().__new__(
            cls, year=cls._value.year, month=cls._value.month, day=cls._value.day
        )

    @classmethod
    def _report_date_or_raise(cls):
        if not cls._cleaned_data.get("report_datetime"):
            raise MedicalDateError(
                {"__all__": "Complete the report date."}, code=MISSING_REPORT_DATETIME
            )

    @classmethod
    def _one_date_or_raise(cls):
        if cls._cleaned_data.get(cls._date_field) and cls._cleaned_data.get(cls._ago_field):
            raise MedicalDateError(
                {
                    cls._ago_field: (
                        "Date conflict. Do not provide a response "
                        f"here if {cls._label} date is available."
                    )
                },
                DATE_AND_AGO_CONFLICT,
            )

    @classmethod
    def _date_or_raise(cls):
        if not cls._cleaned_data.get(cls._date_field) and not cls._cleaned_data.get(
            cls._ago_field
        ):
            raise MedicalDateError(
                {"__all__": f"Complete the {cls._label or '????'} date."},
                code=MISSING_DATE_AND_AGO,
            )

    @classmethod
    def _get_operator(cls):
        msg = f"Is `{cls._field}` supposed to be before or after `{cls._label}` date?"
        if cls._before_reference and cls._after_reference:
            raise MedicalDateError({"__all__": msg}, code=BEFORE_AFTER_BOTH_TRUE)
        if cls._before_reference:
            op = LTE if cls._inclusive else LT
            word = "before"
        elif cls._after_reference:
            op = GTE if cls._inclusive else GT
            word = "after"
        else:
            raise MedicalDateError({"__all__": msg}, code=BEFORE_AFTER_BOTH_FALSE)
        return op, word

    @classmethod
    def _compare_date_and_reference(cls) -> bool:
        value = None
        if cls._op == LT:
            value = cls._value < cls._reference_date
        elif cls._op == LTE:
            value = cls._value <= cls._reference_date
        elif cls._op == GT:
            value = cls._value > cls._reference_date
        elif cls._op == GTE:
            value = cls._value >= cls._reference_date
        return value

    @classmethod
    def _raise_on_failed_comparison(cls):
        formatted_ref = cls._reference_date.strftime(
            convert_php_dateformat(settings.DATE_FORMAT)
        )
        formatted_dte = cls._value.strftime(convert_php_dateformat(settings.DATE_FORMAT))
        inclusive_str = " on or " if cls._inclusive else " "
        raise MedicalDateError(
            {
                cls._field: (
                    f"{cls._label.title()} date must be{inclusive_str}{cls._word} "
                    f"`{formatted_ref}` [{cls._op}]. Got {formatted_dte}"
                )
            },
            code=FAILED_COMPARISON,
        )


class DxDate(MedicalDate):
    def __new__(cls, cleaned_data: dict, **kwargs) -> DxDate:
        defaults = dict(
            before_reference=True,
            reference_date=cleaned_data.get("report_datetime"),
            reference_is_none_msg="Complete the report date first.",
            inclusive=True,
            label="diagnosis",
        )
        defaults.update(**kwargs)
        return super().__new__(cls, "dx_date", "dx_ago", cleaned_data, **defaults)


class RxDate(MedicalDate):
    def __new__(cls, cleaned_data: dict, reference_date: date | DxDate, **kwargs) -> RxDate:
        defaults = dict(
            after_reference=True,
            reference_is_none_msg="Complete the diagnosis date.",
            inclusive=True,
            label="treatment",
        )
        defaults.update(reference_date=reference_date, **kwargs)
        return super().__new__(cls, "rx_init_date", "rx_init_ago", cleaned_data, **defaults)



from django.apps.config import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_model_form"
    verbose_name = "Edc Model Form"
    has_exportable_data = False
    include_in_administration_section = False



from django.apps import apps as django_apps
from django.db import models
from django.db.models import Index

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin

from ..choices import EXPORT_FORMATS
from ..constants import CSV
from ..model_options import ModelOptions


class DataRequest(SiteModelMixin, BaseUuidModel):
    name = models.CharField(max_length=25)

    description = models.TextField(default="")

    decrypt = models.BooleanField(default=False)

    export_format = models.CharField(max_length=25, choices=EXPORT_FORMATS, default=CSV)

    models = models.TextField(
        default="", help_text='List one table per line, no commas. Use "label lower" format.'
    )

    history = HistoricalRecords()

    def __str__(self):
        return self.name

    @property
    def models_as_list(self):
        """Returns `models` as a list of ModelOptions instances.

        Validates each item to be a model name.
        """
        models_as_list = self.models.split("\n")
        models_as_list = [x.strip() for x in models_as_list if x.strip()]
        for model in models_as_list:
            django_apps.get_model(model)
        return [ModelOptions(x) for x in models_as_list]

    class Meta:
        indexes = (Index(fields=["name"]),)



from .account_manager import account_manager
from .administration import administration
from .auditor import auditor
from .celery_manager import celery_manager
from .clinic import clinic
from .clinic_super import clinic_super
from .everyone import everyone



from .age import AgeValueError, age, formatted_age, get_age_in_days, get_dob
from .context_processors_check import edc_context_processors_check
from .dashboard_middleware_check import edc_middleware_check
from .date import ceil_secs, floor_secs, get_utcnow, get_utcnow_as_date, to_local, to_utc
from .disable_signals import DisableSignals
from .get_datetime_from_env import get_datetime_from_env
from .get_static_file import get_static_file
from .get_uuid import get_uuid
from .message_in_queue import message_in_queue
from .round_up import round_half_away_from_zero, round_half_up, round_up
from .show_urls import show_namespaces, show_url_names, show_urls, show_urls_from_patterns
from .text import (
    convert_from_camel,
    convert_php_dateformat,
    escape_braces,
    formatted_date,
    formatted_datetime,
    get_safe_random_string,
    safe_allowed_chars,
    truncate_string,
)

__all__ = [
    "AgeValueError",
    "DisableSignals",
    "age",
    "ceil_secs",
    "convert_from_camel",
    "convert_php_dateformat",
    "edc_context_processors_check",
    "edc_middleware_check",
    "escape_braces",
    "floor_secs",
    "formatted_age",
    "formatted_date",
    "formatted_datetime",
    "get_age_in_days",
    "get_datetime_from_env",
    "get_dob",
    "get_safe_random_string",
    "get_static_file",
    "get_utcnow",
    "get_utcnow_as_date",
    "get_uuid",
    "message_in_queue",
    "paths_for_urlpatterns",
    "round_half_away_from_zero",
    "round_half_up",
    "round_up",
    "safe_allowed_chars",
    "show_namespaces",
    "show_url_names",
    "show_urls",
    "show_urls_from_patterns",
    "to_local",
    "to_utc",
    "truncate_string",
]



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from edc_constants.choices import YES_NO, YES_NO_NA, YES_NO_PENDING_NA
from edc_constants.constants import NOT_APPLICABLE, YES
from edc_lab.choices import VL_QUANTIFIER_NA
from edc_model import estimated_date_from_ago
from edc_model import models as edc_models
from edc_model.validators import date_not_future
from edc_reportable import CELLS_PER_MILLIMETER_CUBED_DISPLAY, COPIES_PER_MILLILITER


class HivArvInitiationModelMixin(models.Model):
    arv_initiated = models.CharField(
        verbose_name="Has the patient started antiretroviral therapy (ART)?",
        max_length=15,
        choices=YES_NO,
        default=YES,
    )

    arv_initiation_ago = edc_models.DurationYMDField(
        verbose_name="How long ago did the patient start ART?",
        null=True,
        blank=True,
    )

    arv_initiation_actual_date = models.DateField(
        verbose_name="Date started antiretroviral therapy (ART)",
        validators=[date_not_future],
        null=True,
        blank=True,
        help_text="Calculated based on response to `arv_initiation_ago`",
    )

    arv_initiation_estimated_date = models.DateField(
        verbose_name="Estimated Date started antiretroviral therapy (ART)",
        validators={date_not_future},
        null=True,
        editable=False,
        help_text="Calculated based on response to `arv_initiation_ago`",
    )

    arv_initiation_date_is_estimated = models.CharField(
        verbose_name="Was the ART start date estimated?",
        max_length=15,
        choices=YES_NO,
        default=YES,
        editable=False,
    )

    def save(self, *args, **kwargs):
        self.dx_estimated_date = estimated_date_from_ago(instance=self, ago_field="dx_ago")
        self.arv_initiation_estimated_date = estimated_date_from_ago(
            instance=self, ago_field="arv_initiation_ago"
        )
        super().save(*args, **kwargs)

    @property
    def best_art_initiation_date(self):
        return self.arv_initiation_actual_date or self.arv_initiation_estimated_date

    class Meta:
        abstract = True


class HivArvMonitoringModelMixin(models.Model):
    # Viral Load
    has_vl = models.CharField(
        verbose_name="Is the patient's most recent viral load result available?",
        max_length=25,
        choices=YES_NO_PENDING_NA,
        default=NOT_APPLICABLE,
    )
    vl = models.IntegerField(
        verbose_name="Most recent viral load",
        validators=[MinValueValidator(0), MaxValueValidator(9999999)],
        null=True,
        blank=True,
        help_text=COPIES_PER_MILLILITER,
    )

    vl_quantifier = models.CharField(
        max_length=10,
        choices=VL_QUANTIFIER_NA,
        null=True,
        blank=True,
    )

    vl_date = models.DateField(
        verbose_name="Date of most recent viral load",
        validators=[date_not_future],
        null=True,
        blank=True,
    )

    # CD4
    has_cd4 = models.CharField(
        verbose_name="Is the patient's most recent CD4 result available?",
        max_length=25,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    cd4 = models.IntegerField(
        verbose_name="Most recent CD4",
        validators=[MinValueValidator(0), MaxValueValidator(3000)],
        null=True,
        blank=True,
        help_text=CELLS_PER_MILLIMETER_CUBED_DISPLAY,
    )

    cd4_date = models.DateField(
        verbose_name="Date of most recent CD4",
        validators=[date_not_future],
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True



def egfr_percent_change(egfr_value: float, baseline_egfr_value: float) -> float:
    """Returns the percent change from baseline"""
    if egfr_value and baseline_egfr_value:
        return 100 * (
            (float(baseline_egfr_value) - float(egfr_value)) / float(baseline_egfr_value)
        )
    return 0.0



# Generated by Django 3.2.13 on 2022-09-06 23:02

import uuid

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0012_auto_20220907_0157"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="refill_identifier",
            field=models.CharField(default=uuid.uuid4, editable=False, max_length=36),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="refill_start_datetime",
            field=models.DateTimeField(
                help_text="Starting date for this refill",
                verbose_name="Refill start date/time",
            ),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="refill_identifier",
            field=models.CharField(default=uuid.uuid4, editable=False, max_length=36),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="refill_start_datetime",
            field=models.DateTimeField(
                help_text="Starting date for this refill",
                verbose_name="Refill start date/time",
            ),
        ),
    ]



import csv
import os
import re
from datetime import datetime

from django.db import models
from django.db.models import Index

from edc_constants.constants import UUID_PATTERN
from edc_model.models import BaseUuidModel

from .file_history import FileHistory


class UploadExportReceiptFileManager(models.Manager):
    def get_by_natural_key(self, file_name):
        return self.get(file_name=file_name)


class UploadExportReceiptFile(BaseUuidModel):
    export_receipt_file = models.FileField(
        upload_to=os.path.join("media", "edc_export", "uploads")
    )

    file_name = models.CharField(max_length=50, null=True, editable=False, unique=True)

    app_label = models.CharField(max_length=50)

    model_name = models.CharField(max_length=50)

    accepted = models.IntegerField(default=0, editable=False)

    duplicate = models.IntegerField(default=0, editable=False)

    total = models.IntegerField(default=0, editable=False)

    errors = models.TextField(editable=False, default="")

    receipt_datetime = models.DateTimeField(editable=False, null=True)

    objects = UploadExportReceiptFileManager()

    def save(self, *args, **kwargs):
        if not self.id:
            self.file_name = self.export_receipt_file.name.replace("\\", "/").split("/")[-1]
            self.update_file_history()
        super(UploadExportReceiptFile, self).save(*args, **kwargs)

    def natural_key(self):
        return (self.file_name,)

    def update_file_history(self):
        """Reads the csv file and updates the export history for
        the given export_uuid.
        """
        self.export_receipt_file.open()
        reader = csv.reader(self.export_receipt_file)
        re_pk = re.compile(UUID_PATTERN)
        error_list = []
        for row in reader:
            self.total += 1
            for item in row:
                if re.match(re_pk, item):  # match a row item on uuid
                    if not FileHistory.objects.filter(export_uuid=item):
                        error_list.append(item)
                    elif FileHistory.objects.filter(export_uuid=item, received=True):
                        self.duplicate += 1
                    else:
                        file_history = FileHistory.objects.get(export_uuid=item)
                        file_history.received = True
                        file_history.received_datetime = datetime.today()
                        file_history.save()
                        self.accepted += 1
        self.errors = "; ".join(error_list)

    class Meta(BaseUuidModel.Meta):
        indexes = (Index(fields=["created"]),)



from __future__ import annotations

from typing import Any

from django.db import models

from edc_constants.choices import YES_NO, YES_NO_NA

from ..calculate_missing import calculate_missing
from ..get_summary import get_summary


class BloodResultsFieldsModelMixin(models.Model):
    results_abnormal = models.CharField(
        verbose_name="Are any of the above results abnormal?",
        choices=YES_NO,
        max_length=25,
        help_text=(
            "Abnormal results present at baseline or continuing from baseline not included."
        ),
    )

    results_reportable = models.CharField(
        verbose_name="If any results are abnormal, are results within grade 3 or above?",
        max_length=25,
        choices=YES_NO_NA,
        help_text=(
            "If YES, this value will open Adverse Event Form. Grade 3 and 4 results "
            "present at baseline or continuing from baseline not included"
        ),
    )

    summary = models.TextField(default="", blank=True)

    reportable_summary = models.TextField(default="", blank=True)

    abnormal_summary = models.TextField(default="", blank=True)

    errors = models.TextField(default="", blank=True)

    missing_count = models.IntegerField(
        default=0,
        editable=False,
        help_text="A count of fields left blank",
    )

    missing = models.TextField(
        default="",
        editable=False,
        help_text="calculated string of field names that have been left blank",
    )

    class Meta:
        abstract = True


class BloodResultsMethodsModelMixin(models.Model):
    """Requires additional attrs `subject_visit` and `requisition`"""

    def save(self, *args, **kwargs):
        reportable, abnormal, errors = self.get_summary()
        self.summary = "\n".join(reportable + abnormal)
        self.reportable_summary = "\n".join(reportable)
        self.abnormal_summary = "\n".join(abnormal)
        self.errors = "\n".join(errors)
        self.missing_count, self.missing = calculate_missing(self, self.lab_panel)
        super().save(*args, **kwargs)

    def get_summary(self):
        return get_summary(self)

    def get_summary_options(self: Any) -> dict:
        # note: gender and dob are queried from RegisteredSubject
        # in reportables
        return dict(
            subject_identifier=self.subject_visit.subject_identifier,
            report_datetime=self.report_datetime,
            age_units="years",
        )

    def get_action_item_reason(self):
        return self.summary

    @property
    def abnormal(self: Any):
        return self.results_abnormal

    @property
    def reportable(self: Any):
        return self.results_reportable

    class Meta:
        abstract = True


class BloodResultsModelMixin(
    BloodResultsFieldsModelMixin, BloodResultsMethodsModelMixin, models.Model
):
    """For each `result` the field name or its prefix should
    match with a value in reportables.

    For example:
        field_name = creatinine_value
        reportables name: creatinine
        value_field_suffix = "_value"

        -OR-

        field_name = creatinine
        reportables name: creatinine
        value_field_suffix = None

    Requires additional attrs `subject_visit` and `requisition`
    from CrfWithRequisitionModelMixin

    """

    class Meta:
        abstract = True



"""
EXTRA_LANG_INFO is a dictionary structure to provide meta information
about 'custom' languages not provided by Django (i.e. those not already
defined in: django.conf.locale.LANG_INFO).

Use 2 digit (639-1) code over 3 digit (639-2) code where available

`code` src: https://www.loc.gov/standards/iso639-2/php/code_list.php
`name_local` src: https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes
"""

EXTRA_LANG_INFO = {
    "lg": {
        "bidi": False,
        "code": "lg",
        "name": "Ganda",
        "name_local": "Luganda",
    },
    "mas": {
        "bidi": False,
        "code": "mas",
        "name": "Maasai",
        "name_local": "ɔl Maa",
    },
    "ry": {
        "bidi": False,
        "code": "ry",
        "name": "Runyakitara",
        "name_local": "Runyakitara",
    },
    "rny": {
        "bidi": False,
        "code": "rny",
        "name": "Runyankore",
        "name_local": "Runyankore",
    },
    "st": {
        "bidi": False,
        "code": "st",
        "name": "Sotho, Southern",
        "name_local": "Sesotho [southern]",
    },
    "tn": {
        "bidi": False,
        "code": "tn",
        "name": "Tswana",
        "name_local": "Setswana",
    },
    "xh": {
        "bidi": False,
        "code": "xh",
        "name": "Xhosa",
        "name_local": "isiXhosa",
    },
    "zu": {
        "bidi": False,
        "code": "zu",
        "name": "Zulu",
        "name_local": "isiZulu",
    },
}



from warnings import warn

import pandas as pd
from django.apps import apps as django_apps

from edc_model_to_dataframe.model_to_dataframe import ModelToDataframe


def missing_subject_identifiers(
    self, model=None, subject_identifiers=None, remove_uuids=None, verbose=None
):
    """Returns a series of subject_identifiers not in model.

    For example:
        missing_subject_identifiers(
            model='edc_registration.registeredsubject',
            subject_identifiers=[a list of subject identifiers])
    """
    # convert list of subject identifiers to a dataframe
    df_subject_identifiers = pd.DataFrame(subject_identifiers, columns=["subject_identifier"])
    df_subject_identifiers["identifier"] = df_subject_identifiers["subject_identifier"]
    df_subject_identifiers = df_subject_identifiers.set_index("subject_identifier")
    if verbose:
        df_subject_identifiers.head()

    # load model into dataframe
    model_cls = django_apps.get_model(model)
    df_subject = ModelToDataframe(
        model_cls.objects.values("subject_identifier", "gender", "dob").all()
    ).dataframe
    # drop duplicates
    df_subject = df_subject.drop_duplicates()
    # remove subject identifier as UUID
    if remove_uuids:
        df_subject = df_subject[df_subject["subject_identifier"].str.len() != 32]
    # set index to subject_identifier
    df_subject = df_subject.set_index("subject_identifier")
    # filter df of subject identifiers to leave only those not in model
    df_missing = df_subject_identifiers[
        -df_subject_identifiers["identifier"].isin(df_subject.index)
    ]

    if len(df_missing.index) > 0 and verbose:
        warn(
            f"There are {len(df_missing['identifier'])} subject identifiers "
            f"missing from {model}."
        )
    return df_missing["identifier"]



COMPLETED_FOLLOWUP = "completed_followup"
CONSENT_WITHDRAWAL = "consent_withdrawal"
END_OF_STUDY_ACTION = "submit_end_of_study"
INVALID_ENROLMENT = "invalid_enrolment"
LATE_EXCLUSION = "late_exclusion"
OTHER_RX_DISCONTINUATION = "other_treatment_discontinuation"
WITHDRAWAL = "withdrawal"



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.views.generic import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin


@method_decorator(login_required, name="dispatch")
class ReturnView(EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView):
    model_pks: list[str] | None = None
    template_name: str = "edc_pharmacy/stock/return.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import TYPE_CHECKING
from uuid import uuid4

from django.utils.translation import gettext as _

from edc_model.utils import get_history_url

if TYPE_CHECKING:
    from edc_metadata.models import CrfMetadata

__all__ = ["HistoryButton"]


@dataclass
class HistoryButton:
    model_obj: InitVar[CrfMetadata] = None
    color: str = field(default="default")
    url: str = field(default=None, init=False)
    disabled: str = field(default="disabled", init=False)
    btn_id: str = field(init=False)

    def __post_init__(self, model_obj: CrfMetadata):
        if model_obj.model_instance:
            self.url = get_history_url(model_obj.model_instance)
            if self.url:
                self.disabled = ""
                self.btn_id = (
                    f"{model_obj.model_cls._meta.label_lower.split('.')[1]}-history-"
                    f"{model_obj.model_instance.id.hex}"
                )
        else:
            self.btn_id = (
                f"{model_obj.model_cls._meta.label_lower.split('.')[1]}-history-"
                f"{uuid4().hex}"
            )

    @property
    def label(self) -> str:
        return _("Audit")



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from edc_utils import formatted_datetime, to_utc
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from .exceptions import OffstudyError

if TYPE_CHECKING:
    from django.db.models import Model

    from .model_mixins import OffstudyModelMixin


def get_offstudy_model() -> str:
    """Returns the Offstudy model name in label_lower format"""
    return site_visit_schedules.get_offstudy_model()


def get_offstudy_model_cls() -> OffstudyModelMixin:
    """Returns the Offstudy model class.

    Uses visit_schedule_name to get the class from the visit schedule
    otherwise defaults settings.EDC_OFFSTUDY_OFFSTUDY_MODEL.
    """
    return django_apps.get_model(site_visit_schedules.get_offstudy_model())


def raise_if_offstudy(
    source_obj: Model | None = None,
    subject_identifier: str = None,
    report_datetime: datetime = None,
) -> OffstudyModelMixin | None:
    """Returns None or raises OffstudyError"""
    obj = None
    try:
        with transaction.atomic():
            obj = get_offstudy_model_cls().objects.get(
                subject_identifier=subject_identifier,
                offstudy_datetime__lt=to_utc(report_datetime),
            )
    except ObjectDoesNotExist:
        pass
    else:
        msg_part = f"Source model `{source_obj._meta.verbose_name}`." if source_obj else ""
        raise OffstudyError(
            "Subject off study by given date/time. "
            f"Got report_datetime=`{formatted_datetime(report_datetime)}` "
            f"while the offstudy date is `{formatted_datetime(obj.offstudy_datetime)}` "
            f"Subject {subject_identifier}. {msg_part} "
            f"See also '{get_offstudy_model_cls()._meta.verbose_name}'."
        )
    return obj



from __future__ import annotations

from django.contrib.admin import SimpleListFilter
from django.core.exceptions import FieldError
from django.db.models import Count, Q

from edc_constants.constants import OTHER
from edc_sites.site import sites


class ListFieldWithOtherListFilter(SimpleListFilter):
    title: str
    parameter_name: str
    other_parameter_name: str

    def lookups(self, request, model_admin):
        values_list = []
        try:
            values_list = (
                model_admin.model.objects.filter(
                    site_id__in=sites.get_site_ids_for_user(request=request)
                )
                .order_by(f"{self.parameter_name}__name")
                .values_list(
                    f"{self.parameter_name}__name",
                    f"{self.parameter_name}__display_name",
                )
                .annotate(count=Count(f"{self.parameter_name}__name"))
            )
        except (AttributeError, FieldError):
            values_list = (
                model_admin.model.objects.filter(
                    site_id__in=sites.get_site_ids_for_user(request=request)
                )
                .order_by(self.parameter_name)
                .values_list(self.parameter_name)
                .annotate(count=Count(self.parameter_name))
            )
        finally:
            try:
                names = [(value[0], value[1]) for value in values_list if value[2] > 0]
            except IndexError:
                names = [(value[0], value[0]) for value in values_list if value[1] > 0]
        if [n[0] for n in names if n[0] == OTHER]:
            values_list = (
                model_admin.model.objects.filter(
                    site_id__in=sites.get_site_ids_for_user(request=request)
                )
                .order_by(self.other_parameter_name)
                .values_list(self.other_parameter_name)
                .annotate(count=Count(self.other_parameter_name))
            )
            other_names = [
                (f"Other: {value[0]}", f"Other: {value[0].capitalize()[:25]}")
                for value in values_list
                if value[0]
            ]
            names = sorted(names)
            other_names = sorted(other_names)
            names.extend(other_names)
        return tuple(names)

    def queryset(self, request, queryset):
        if self.value() and self.value() != "none":
            value = self.value().replace("Other: ", "")
            try:
                queryset = queryset.filter(
                    (
                        Q(**{f"{self.parameter_name}__name": value})
                        | Q(**{self.other_parameter_name: value})
                    ),
                    site_id__in=sites.get_site_ids_for_user(request=request),
                )
            except AttributeError:
                opts = {
                    "site_id__in": sites.get_site_ids_for_user(request=request),
                    self.parameter_name: value,
                }
                queryset = queryset.filter(**opts)
        return queryset



from django.db import models
from django.db.models import PROTECT
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords

from ...exceptions import LotError
from ..medication import Assignment
from .product import Product


class Manager(models.Manager):
    use_in_migrations = True


class Lot(BaseUuidModel):
    lot_identifier = models.CharField(  # noqa: DJ001
        max_length=25,
        unique=True,
        null=True,
        blank=False,
        help_text="A sequential unique identifier set by the EDC",
    )

    lot_no = models.CharField(verbose_name="Batch", max_length=50, unique=True)

    product = models.ForeignKey(Product, on_delete=PROTECT, null=True, blank=False)

    assignment = models.ForeignKey(Assignment, on_delete=models.PROTECT, null=True, blank=True)

    manufactured_date = models.DateField(null=True, blank=True)

    processed_until_date = models.DateField(null=True, blank=True)

    expiration_date = models.DateField()

    manufactured_by = models.CharField(max_length=50, default="", blank=True)

    country_of_origin = models.CharField(max_length=50, default="", blank=True)

    storage_conditions = models.CharField(max_length=50, default="", blank=True)

    reference = models.CharField(max_length=50, default="", blank=True)

    comment = models.CharField(max_length=250, default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.lot_no}: {self.product.name} "

    def save(self, *args, **kwargs):
        if not self.lot_identifier:
            self.lot_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        if self.assignment != self.product.assignment:
            raise LotError("Assignment mismatch.")
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Lot"
        verbose_name_plural = "Lots"



export_codenames = [
    "edc_export.display_export_admin_action",
    "edc_export.add_datarequest",
    "edc_export.add_datarequesthistory",
    "edc_export.add_exportreceipt",
    "edc_export.add_filehistory",
    "edc_export.add_objecthistory",
    "edc_export.add_plan",
    "edc_export.add_uploadexportreceiptfile",
    "edc_export.change_datarequest",
    "edc_export.change_datarequesthistory",
    "edc_export.change_exportreceipt",
    "edc_export.change_filehistory",
    "edc_export.change_objecthistory",
    "edc_export.change_plan",
    "edc_export.change_uploadexportreceiptfile",
    "edc_export.delete_datarequest",
    "edc_export.delete_datarequesthistory",
    "edc_export.delete_exportreceipt",
    "edc_export.delete_filehistory",
    "edc_export.delete_objecthistory",
    "edc_export.delete_plan",
    "edc_export.delete_uploadexportreceiptfile",
    "edc_export.view_datarequest",
    "edc_export.view_datarequesthistory",
    "edc_export.view_exportreceipt",
    "edc_export.view_filehistory",
    "edc_export.view_historicaldatarequest",
    "edc_export.view_historicalexportreceipt",
    "edc_export.view_historicalfilehistory",
    "edc_export.view_historicalobjecthistory",
    "edc_export.view_historicalplan",
    "edc_export.view_objecthistory",
    "edc_export.view_plan",
    "edc_export.view_uploadexportreceiptfile",
    "edc_export.export_subjectschedulehistory",
    "edc_export.export_visitschedule",
    "edc_export.view_export_dashboard",
]



from django import forms
from django.apps import apps as django_apps

from edc_form_validators import FormValidator, FormValidatorMixin

from ..models import DataRequest


class DataRequestFormValidator(FormValidator):
    pass


class DataRequestForm(FormValidatorMixin, forms.ModelForm):
    form_validator_cls = DataRequestFormValidator

    def clean(self):
        cleaned_data = super().clean()
        if self.cleaned_data.get("requested"):
            requested = self.cleaned_data.get("requested").split("\n")
            requested = [x.strip() for x in requested if x.strip()]
            for model in requested:
                try:
                    django_apps.get_model(model)
                except (LookupError, ValueError):
                    raise forms.ValidationError(f"Invalid model. Got '{model}'")
        return cleaned_data

    class Meta:
        model = DataRequest
        fields = "__all__"



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from edc_reportable import GRAMS_PER_DECILITER, MICROMOLES_PER_LITER
from edc_reportable.units import (
    GRAMS_PER_LITER,
    IU_LITER,
    IU_LITER_DISPLAY,
    MILLIGRAMS_PER_DECILITER,
    MILLIGRAMS_PER_LITER,
)

from ..model_mixin_factories import reportable_result_model_mixin_factory


class AlbuminModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="albumin",
        verbose_name="Serum albumin",
        units_choices=(
            (GRAMS_PER_DECILITER, GRAMS_PER_DECILITER),
            (GRAMS_PER_LITER, GRAMS_PER_LITER),
        ),
        decimal_places=1,
        validators=[MinValueValidator(1.0), MaxValueValidator(9999.9)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class AlpModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="alp",
        units_choices=((IU_LITER, IU_LITER_DISPLAY),),
        decimal_places=2,
        validators=[MinValueValidator(1.00), MaxValueValidator(9999.99)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class AltModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="alt",
        units_choices=((IU_LITER, IU_LITER_DISPLAY),),
        decimal_places=2,
        validators=[MinValueValidator(1.00), MaxValueValidator(9999.99)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class AstModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="ast",
        units_choices=((IU_LITER, IU_LITER_DISPLAY),),
        decimal_places=0,
        validators=[MinValueValidator(1), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class AmylaseModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="amylase",
        verbose_name="Serum Amylase",
        units_choices=((IU_LITER, IU_LITER_DISPLAY),),
        decimal_places=1,
        validators=[MinValueValidator(1.0), MaxValueValidator(9999.9)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class TotalBilirubinModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="tbil",
        verbose_name="Total Bilirubin",
        units_choices=(
            (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
            (GRAMS_PER_LITER, GRAMS_PER_LITER),
            (MICROMOLES_PER_LITER, MICROMOLES_PER_LITER),
        ),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class CrpModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="crp",
        verbose_name="C-reactive protein",
        units_choices=(
            (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
            (MILLIGRAMS_PER_LITER, MILLIGRAMS_PER_LITER),
        ),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class GgtModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="ggt",
        units_choices=((IU_LITER, IU_LITER_DISPLAY),),
        decimal_places=0,
        validators=[MinValueValidator(1), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True



from decimal import Decimal

from django import forms
from django.db.models import Sum

from ...models import ReceiveItem


class ReceiveItemForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()
        if not self.instance.id:
            if not cleaned_data.get("receive"):
                raise forms.ValidationError({"receive": "This field is required"})
            if not cleaned_data.get("order_item"):
                raise forms.ValidationError({"order_item": "This field is required"})
            if not cleaned_data.get("container"):
                raise forms.ValidationError({"container": "This field is required"})

        if cleaned_data.get("order_item") and cleaned_data.get("lot"):
            if (
                cleaned_data.get("order_item").product.assignment
                != cleaned_data.get("lot").assignment
            ):
                raise forms.ValidationError({"lot": "Lot assignment does not match product"})

        # in unit_qty's
        if cleaned_data.get("qty"):
            # TODO: clean this up
            qty_ordered = self.order_item.unit_qty + (self.order_item.unit_qty_received or 0)
            qty_already_received = self._meta.model.objects.filter(
                order_item=self.order_item
            ).aggregate(unit_qty=Sum("unit_qty"))["unit_qty"] or Decimal(0)
            qty_available = qty_ordered - qty_already_received
            qty_to_receive = cleaned_data.get("qty") * cleaned_data.get("container").qty
            if qty_to_receive > qty_available:
                raise forms.ValidationError(
                    {
                        "qty": (
                            f"Exceeds `unit qty` ordered of "
                            f"{self.order_item.unit_qty}. "
                            f"Got {qty_to_receive} but only "
                            f"{qty_available} are still on order."
                        )
                    }
                )

        return cleaned_data

    @property
    def order_item(self):
        return self.cleaned_data.get("order_item") or self.instance.order_item

    class Meta:
        model = ReceiveItem
        fields = "__all__"



from django.apps import apps as django_apps

from .navbar import Navbar
from .navbar_item import NavbarItem
from .site_navbars import site_navbars

app_config = django_apps.get_app_config("edc_navbar")

if app_config.register_default_navbar:
    default_navbar = Navbar(name=app_config.default_navbar_name)

    default_navbar.register(
        NavbarItem(
            name="home",
            title="Home",
            fa_icon="fa-home",
            url_name="home_url",
            codename="edc_navbar.nav_home",
        )
    )

    default_navbar.register(
        NavbarItem(
            name="administration",
            title="Administration",
            fa_icon="fa-cog",
            codename="edc_navbar.nav_administration",
            url_name="administration_url",
        )
    )

    # remove to use a form. See edc_navbar.html
    # default_navbar.register(
    #     NavbarItem(
    #         name="logout",
    #         title="Logout",
    #         fa_icon="fa-sign-out-alt",
    #         url_name="edc_auth_admin:logout",
    #         codename="edc_navbar.nav_logout",
    #     )
    # )

    site_navbars.register(default_navbar)



from __future__ import annotations

import csv
import sys
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.utils import timezone
from tqdm import tqdm

from edc_sites.site import sites

from .exceptions import HolidayFileNotFoundError, HolidayImportError
from .utils import get_holiday_model_cls

if TYPE_CHECKING:
    from .models import Holiday

LOCAL_DATE = 0
LABEL = 1
COUNTRY = 2


def import_holidays(verbose: bool | None = None, test: bool | None = None) -> None:
    model_cls = get_holiday_model_cls()
    if test:
        import_for_tests(model_cls)
    else:
        path = Path(settings.HOLIDAY_FILE)
        try:
            if not Path(path).exists():
                raise HolidayFileNotFoundError(path)
        except TypeError as e:
            raise HolidayImportError(
                f"Invalid path importing holiday file. See settings.HOLIDAY_FILE. Got {path}."
            ) from e
        if verbose:
            sys.stdout.write(
                f"\nImporting holidays from '{path}' into {model_cls._meta.label_lower}\n"
            )
        model_cls.objects.all().delete()

        recs = check_for_duplicates_in_file(path)

        import_file(path, recs, model_cls)

        if verbose:
            sys.stdout.write("Done.\n")


def check_for_duplicates_in_file(path: Path) -> list:
    """Returns a list of records."""
    with path.open("r") as f:
        reader = csv.DictReader(f, fieldnames=["local_date", "label", "country"])
        recs = [(row["local_date"], row["country"]) for row in reader]
    if len(recs) != len(list(set(recs))):
        raise HolidayImportError("Invalid file. Duplicate dates detected for a country")
    return recs


def import_file(path: Path, recs: list, model_cls: Holiday):
    with path.open("r") as f:
        reader = csv.DictReader(f, fieldnames=["local_date", "label", "country"])
        for index, row in tqdm(enumerate(reader), total=len(recs)):
            if index == 0:
                continue
            try:
                local_date = datetime.strptime(row["local_date"], "%Y-%m-%d").date()  # noqa: DTZ007
            except ValueError as e:
                raise HolidayImportError(
                    f"Invalid format when importing from {path}. Got '{e}'"
                ) from e
            else:
                try:
                    obj = model_cls.objects.get(country=row["country"], local_date=local_date)
                except ObjectDoesNotExist:
                    model_cls.objects.create(
                        country=row["country"], local_date=local_date, name=row["label"]
                    )
                else:
                    obj.name = row["label"]
                    obj.save()


def import_for_tests(model_cls: type[Holiday]):
    year = timezone.now().year
    country = sites.get_current_country()
    if not country:
        raise HolidayImportError(
            "Cannot determine default country when importing "
            "holidays for tests. Confirm SITE_ID is valid. See `import_for_tests`"
        )
    rows = [
        [f"{year}-01-02", "Public Holiday", country],
        [f"{year}-01-01", "New Year", country],
        [f"{year}-04-14", "Good Friday", country],
        [f"{year}-04-17", "Easter Monday", country],
        [f"{year}-05-01", "May Day/Labour Day", country],
        [f"{year}-05-25", "Ascension Day", country],
        [f"{year}-07-18", "Public Holiday", country],
        [f"{year}-09-30", "Botswana Day", country],
        [f"{year}-10-02", "Public Holiday", country],
        [f"{year}-12-25", "Christmas Day", country],
        [f"{year}-12-26", "Boxing Day", country],
    ]
    objs = []
    for index, row in tqdm(enumerate(rows), total=len(rows)):
        if index == 0:
            continue
        try:
            local_date = datetime.strptime(row[LOCAL_DATE], "%Y-%m-%d").date()  # noqa: DTZ007
        except ValueError as e:
            raise HolidayImportError(
                f"Invalid format when importing holidays (test). Got '{e}'"
            ) from e
        else:
            objs.append(
                model_cls(country=row[COUNTRY], local_date=local_date, name=row[LABEL])
            )
    with transaction.atomic():
        model_cls.objects.all().delete()
        model_cls.objects.bulk_create(objs)



import pandas as pd

from edc_constants.constants import FEMALE, MALE

from ..constants import MEDIAN_IQR, N_ONLY, N_WITH_COL_PROP, N_WITH_ROW_PROP
from ..row import RowDefinition, RowDefinitions
from ..table import Table


class FastingFbgTable(Table):
    def __init__(
        self,
        main_df: pd.DataFrame = None,
        colname: str | None = None,
        title: str | None = None,
    ):
        colname = colname or "fasting_fbg_hrs"
        title = title or "Fasting duration (hrs)"
        super().__init__(colname=colname, main_df=main_df, title=title)

    @property
    def row_definitions(self) -> RowDefinitions:
        df_tmp = self.main_df.copy()
        row_defs = RowDefinitions(reverse_rows=False)
        row0 = RowDefinition(
            title=self.title,
            label=self.default_sublabel,
            condition=(df_tmp["gender"].notna()),
            columns={FEMALE: (N_ONLY, 2), MALE: (N_ONLY, 2), "All": (N_ONLY, 2)},
            drop=False,
        )
        row_defs.add(row0)
        columns = {
            FEMALE: (MEDIAN_IQR, 2),
            MALE: (MEDIAN_IQR, 2),
            "All": (MEDIAN_IQR, 2),
        }
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG fasted (hours), median (IQR)",
                condition=(self.main_df[self.colname].notna()),
                columns=columns,
                drop=False,
            )
        )
        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG Fasted <8.0 hrs",
                condition=(self.main_df[self.colname] < 8.0),
                columns=columns,
                drop=False,
            )
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="not measured",
                condition=(self.main_df[self.colname].isna()),
                columns=columns,
                drop=False,
            )
        )
        return row_defs


class FastingOgttTable(Table):
    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(colname="fasting_ogtt_hrs", main_df=main_df, title="")

    @property
    def row_definitions(self) -> RowDefinitions:
        row_defs = RowDefinitions(reverse_rows=False)
        columns = {
            FEMALE: (MEDIAN_IQR, 2),
            MALE: (MEDIAN_IQR, 2),
            "All": (MEDIAN_IQR, 2),
        }
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT fasted (hours), median (IQR)",
                condition=(self.main_df[self.colname].notna()),
                columns=columns,
                drop=False,
            )
        )
        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT Fasted <8.0 hrs",
                condition=(self.main_df[self.colname] < 8.0),
                columns=columns,
                drop=False,
            )
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="not measured",
                condition=(self.main_df[self.colname].isna()),
                columns=columns,
                drop=False,
            )
        )
        return row_defs


class FastingTable(Table):
    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(main_df=main_df, title="Fasting duration (hrs)")

    def build_table_df(self) -> None:
        df1 = FastingFbgTable(main_df=self.main_df).table_df
        df2 = FastingOgttTable(main_df=self.main_df).table_df
        self.table_df = pd.concat([df1, df2])



import pandas as pd
from django.apps import apps as django_apps
from django.db import models


def refresh_model_from_dataframe(
    df: pd.DataFrame,
    model: str | None = None,
    model_cls: type[models.Model] | None = None,
    columns: list[str] | None = None,
) -> None:
    columns = columns or list(df.columns)
    model_cls = model_cls or django_apps.get_model(model)
    model_cls.objects.all().delete()
    model_cls.objects.bulk_create(
        [model_cls(**{col: row[col] for col in columns}) for _, row in df.iterrows()]
    )



from .blood_pressure import DiastolicPressureField, SystolicPressureField
from .heart_rate import HeartRateField
from .height import HeightField
from .respiratory_rate import RespiratoryRateField
from .temperature import TemperatureField
from .waist_circumference import WaistCircumferenceField
from .weight import WeightField



# Generated by Django 5.2.3 on 2025-06-27 17:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0073_alter_stockout_options"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="StockOut",
            new_name="StockAvailability",
        ),
        migrations.RenameIndex(
            model_name="stockavailability",
            new_name="edc_pharmac_subject_3f9cb2_idx",
            old_name="edc_pharmac_subject_63b1ce_idx",
        ),
    ]



from django import forms

from edc_action_item.forms import ActionItemFormMixin
from edc_form_validators import FormValidatorMixin
from edc_model_form.mixins import BaseModelFormMixin
from edc_offstudy.modelform_mixins import OffstudyNonCrfModelFormMixin
from edc_sites.forms import SiteModelFormMixin

from ..form_validators import ProtocolDeviationViolationFormValidator
from ..models import ProtocolDeviationViolation


class ProtocolDeviationViolationForm(
    SiteModelFormMixin,
    OffstudyNonCrfModelFormMixin,
    FormValidatorMixin,
    ActionItemFormMixin,
    BaseModelFormMixin,
    forms.ModelForm,
):
    report_datetime_field_attr = "report_datetime"

    form_validator_cls = ProtocolDeviationViolationFormValidator

    subject_identifier = forms.CharField(
        label="Subject Identifier",
        required=False,
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    class Meta:
        model = ProtocolDeviationViolation
        fields = "__all__"



from edc_constants.constants import NOT_APPLICABLE, OTHER

from .constants import PER_DAY, PER_HR, SINGLE

list_data = {
    "edc_pharmacy.formulationtype": [
        ("tablet", "Tablet"),
        ("capsule", "Capsule"),
        ("vial", "Vial"),
        ("liquid", "Liquid"),
        ("powder", "Powder"),
        ("suspension", "Suspension"),
        ("gel", "Gel"),
        ("oil", "Oil"),
        ("lotion", "Lotion"),
        ("cream", "Cream"),
        ("patch", "Patch"),
        (OTHER, "Other"),
    ],
    "edc_pharmacy.units": [
        ("mg", "mg"),
        ("ml", "ml"),
        ("g", "g"),
        (OTHER, "Other ..."),
        (NOT_APPLICABLE, "Not applicable"),
    ],
    "edc_pharmacy.route": [
        ("intramuscular", "Intramuscular"),
        ("intravenous", "Intravenous"),
        ("oral", "Oral"),
        ("topical", "Topical"),
        ("subcutaneous", "Subcutaneous"),
        ("intravaginal", "Intravaginal"),
        ("rectal", "Rectal"),
        (OTHER, "Other"),
    ],
    "edc_pharmacy.frequencyunits": [
        (PER_HR, "times per hour"),
        (PER_DAY, "times per day"),
        (SINGLE, "single dose"),
        (OTHER, "Other ..."),
        (NOT_APPLICABLE, "Not applicable"),
    ],
    "edc_pharmacy.containertype": [
        ("tablet", "Tablet"),
        ("bottle", "Bottle"),
    ],
    "edc_pharmacy.containerunits": [
        ("tablet", "Tablet"),
    ],
    "edc_pharmacy.location": [
        ("central", "Central"),
    ],
}



# Generated by Django 4.0.4 on 2022-06-02 01:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0029_appointment_ignore_window_period_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_status",
            field=models.CharField(
                choices=[
                    ("new", "Not started"),
                    ("in_progress", "In Progress"),
                    ("incomplete", "Incomplete"),
                    ("done", "Done"),
                    ("cancelled", "Cancelled"),
                ],
                db_index=True,
                default="new",
                help_text="If the visit has already begun, only 'in progress', 'incomplete' or 'done' are valid options. Only unscheduled appointments may be cancelled.",
                max_length=25,
                verbose_name="Status",
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="appt_timing",
            field=models.CharField(
                choices=[
                    ("ontime", "On time (within window period)"),
                    ("missed", "Missed"),
                ],
                default="ontime",
                help_text="If late, you may be required to complete a Protocol Deviation / Violation form. Refer to the protocol/SOP for the definition of scheduled appointment window periods.",
                max_length=25,
                verbose_name="Timing",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_status",
            field=models.CharField(
                choices=[
                    ("new", "Not started"),
                    ("in_progress", "In Progress"),
                    ("incomplete", "Incomplete"),
                    ("done", "Done"),
                    ("cancelled", "Cancelled"),
                ],
                db_index=True,
                default="new",
                help_text="If the visit has already begun, only 'in progress', 'incomplete' or 'done' are valid options. Only unscheduled appointments may be cancelled.",
                max_length=25,
                verbose_name="Status",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_timing",
            field=models.CharField(
                choices=[
                    ("ontime", "On time (within window period)"),
                    ("missed", "Missed"),
                ],
                default="ontime",
                help_text="If late, you may be required to complete a Protocol Deviation / Violation form. Refer to the protocol/SOP for the definition of scheduled appointment window periods.",
                max_length=25,
                verbose_name="Timing",
            ),
        ),
    ]



from django.contrib import admin

from edc_list_data.admin import ListModelAdminMixin

from ...admin_site import edc_pharmacy_admin
from ...models import FrequencyUnits


@admin.register(FrequencyUnits, site=edc_pharmacy_admin)
class FrequencyUnitsAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass



# Generated by Django 3.2.13 on 2022-07-14 18:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_offstudy", "0010_auto_20220704_1841"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectoffstudy",
            name="offstudy_reason",
            field=models.CharField(
                choices=[
                    ("lost", "Lost to follow-up"),
                    ("completed protocol", "Completed protocol"),
                    ("consent_withdrawn", "Consent withdrawn"),
                    ("dead", "Deceased"),
                ],
                max_length=125,
                verbose_name="Please code the primary reason participant taken off-study",
            ),
        ),
        migrations.AlterField(
            model_name="subjectoffstudy",
            name="offstudy_reason",
            field=models.CharField(
                choices=[
                    ("lost", "Lost to follow-up"),
                    ("completed protocol", "Completed protocol"),
                    ("consent_withdrawn", "Consent withdrawn"),
                    ("dead", "Deceased"),
                ],
                max_length=125,
                verbose_name="Please code the primary reason participant taken off-study",
            ),
        ),
    ]



from django.contrib import admin
from django.db.models import QuerySet
from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html
from rangefilter.filters import DateRangeFilterBuilder, NumericRangeFilterBuilder

from edc_model_admin.dashboard import ModelAdminDashboardMixin
from edc_model_admin.mixins import TemplatesModelAdminMixin
from edc_qareports.modeladmin_mixins import QaReportModelAdminMixin
from edc_sites.admin import SiteModelAdminMixin

from ...admin_site import edc_pharmacy_admin
from ...analytics.dataframes.no_stock_for_subjects_df import stock_for_subjects_df
from ...models import StockAvailability
from ..list_filters import HasCodesListFilter


def wrap_html(s, url):
    return ",".join([f"<a href='{url}?q={c}'>{c}</a>" for c in s.split(",")])


def update_report(modeladmin, request):
    now = timezone.now()
    created = 0
    modeladmin.model.objects.all().delete()

    df = stock_for_subjects_df()
    url = reverse("edc_pharmacy_admin:edc_pharmacy_storagebinitem_changelist")
    df.loc[~df["codes"].isna(), "codes"] = df.loc[~df["codes"].isna(), "codes"].apply(
        lambda s: wrap_html(s, url)
    )
    df.loc[df["codes"].isna(), "codes"] = None

    df.loc[~df["bins"].isna(), "bins"] = df.loc[~df["bins"].isna(), "bins"].apply(
        lambda s: wrap_html(s, url)
    )
    df.loc[df["bins"].isna(), "bins"] = None
    if not df.empty:
        data = [
            modeladmin.model(
                subject_identifier=row["subject_identifier"],
                site_id=row["site_id"],
                visit_code=row["visit_code"],
                appt_date=row["appt_date"],
                relative_days=row["relative_days"],
                codes=row["codes"],
                bins=row["bins"],
                report_model=modeladmin.model._meta.label_lower,
                created=now,
            )
            for _, row in df.iterrows()
        ]
        return len(modeladmin.model.objects.bulk_create(data))
    return created


@admin.register(StockAvailability, site=edc_pharmacy_admin)
class StockAvailabilityModelAdmin(
    QaReportModelAdminMixin,
    SiteModelAdminMixin,
    ModelAdminDashboardMixin,
    TemplatesModelAdminMixin,
    admin.ModelAdmin,
):
    queryset_filter: dict | None = None
    qa_report_list_display_insert_pos = 3
    include_note_column = False
    ordering = ["relative_days"]
    list_display = [
        "dashboard",
        "subject",
        "site",
        "visit_code",
        "appt_date",
        "relative_days",
        "formatted_codes",
        "formatted_bins",
        "last_updated",
    ]

    list_filter = [
        HasCodesListFilter,
        ("appt_date", DateRangeFilterBuilder()),
        ("relative_days", NumericRangeFilterBuilder()),
        "visit_code",
        "site_id",
    ]

    search_fields = ["subject_identifier"]

    def get_queryset(self, request) -> QuerySet:
        update_report(self, request)
        qs = super().get_queryset(request)
        if self.queryset_filter:
            qs = qs.filter(**self.queryset_filter)
        return qs

    @admin.display(description="Codes", ordering="codes")
    def formatted_codes(self, obj):
        if obj.codes:
            return format_html(f'<span style="font-family:courier;">{obj.codes}</span>')
        return None

    @admin.display(description="Bins", ordering="bins")
    def formatted_bins(self, obj):
        if obj.codes:
            return format_html(f'<span style="font-family:courier;">{obj.bins}</span>')
        return None

    @admin.display(description="subject", ordering="subject_identifier")
    def subject(self, obj=None):
        return obj.subject_identifier

    @admin.display(description="visit", ordering="visit_code")
    def visit(self, obj=None):
        return obj.visit_code

    @admin.display(description="last_updated", ordering="created")
    def last_updated(self, obj=None):
        return obj.created

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in request.user.userprofile.sites.all() if s.id != request.site.id]



from django.db import models


class AddressMixin(models.Model):
    contact_name = models.CharField(max_length=50, default="", blank=True)

    address = models.CharField(verbose_name="Address", max_length=50)

    postal_code = models.CharField(default="0000", max_length=50)

    city = models.CharField(max_length=50)

    state = models.CharField(
        verbose_name="State or Province", default="", blank=True, max_length=50
    )

    country = models.CharField(max_length=50)

    telephone = models.CharField(default="", blank=True, max_length=50)

    mobile = models.CharField(default="", blank=True, max_length=50)

    fax = models.CharField(default="", blank=True, max_length=50)

    email = models.EmailField(default="", blank=True, max_length=50)

    class Meta:
        abstract = True



from edc_reportable.exceptions import NotEvaluated
from edc_reportable.utils import get_reference_range_collection

__all__ = ["get_summary"]


def get_summary(obj) -> tuple[list[str], list[str], list[str]]:
    """Returns a list of graded or abnormal values given a
    BloodResultsModel instance.
    """
    opts = obj.get_summary_options()
    reportable = []
    abnormal = []
    errors = []
    for field_name in [f.name for f in obj._meta.get_fields()]:
        try:
            utest_id, _ = field_name.split("_value")
        except ValueError:
            utest_id = field_name
        reference_range_collection = get_reference_range_collection(obj)
        units = getattr(obj, f"{utest_id}_units", None)
        value = getattr(obj, field_name)
        if units and value:
            opts.update(units=units, label=utest_id)
            try:
                grading_data, grading_eval_phrase = reference_range_collection.get_grade(
                    value, **opts
                )
            except NotEvaluated as e:
                errors.append(f"{e}.")
                grading_data = None
                grading_eval_phrase = None
            if (
                grading_data
                and grading_data.grade
                in reference_range_collection.reportable_grades(utest_id)
            ):
                setattr(obj, f"{utest_id}_grade", grading_data.grade)
                setattr(obj, f"{utest_id}_grade_description", grading_data.description)
                reportable.append(f"{grading_eval_phrase}")
            else:
                try:
                    is_normal, normal_data = reference_range_collection.is_normal(
                        value, **opts
                    )
                except NotEvaluated as e:
                    errors.append(f"{e}.")
                else:
                    if is_normal is False:
                        abnormal.append(
                            f"{normal_data.label}: {value} {units} "
                            f"{normal_data.phrase} Abnormal"
                        )
    return reportable, abnormal, errors



from typing import Any

from django.urls.base import reverse
from django.views.generic.base import ContextMixin

from edc_dashboard.url_names import url_names


class FormActionViewError(Exception):
    pass


class FormActionViewMixin(ContextMixin):
    action_name = None
    form_action_name = "form_action"
    form_action_selected_items_name = "selected_items"
    form_action_url = None

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            action_name=self.action_name,
            form_action_name=self.form_action_name,
            form_action_selected_items_name=self.form_action_selected_items_name,
            form_action_url_reversed=self.form_action_url_reversed,
        )
        return super().get_context_data(**kwargs)

    @property
    def form_action_url_kwargs(self):
        return self.url_kwargs

    @property
    def form_action_url_reversed(self):
        form_action_url = url_names.get(self.form_action_url)
        return reverse(form_action_url, kwargs=self.form_action_url_kwargs)



from django import forms

from edc_form_validators import FormValidatorMixin

from ..form_validators import HealthFacilityFormValidator
from ..models import HealthFacility


class HealthFacilityForm(FormValidatorMixin, forms.ModelForm):
    form_validator_cls = HealthFacilityFormValidator

    class Meta:
        model = HealthFacility
        fields = "__all__"



# Generated by Django 3.2.13 on 2022-07-08 05:05

import uuid

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models

import edc_pharmacy.models.medication


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("sites", "0002_alter_domain_unique"),
        ("edc_pharmacy", "0008_auto_20220706_2324"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="HistoricalMedicationProduct",
            new_name="HistoricalProduct",
        ),
        migrations.RenameModel(
            old_name="MedicationProduct",
            new_name="Product",
        ),
        migrations.RenameModel(
            old_name="HistoricalMedicationStock",
            new_name="HistoricalStock",
        ),
        migrations.RenameModel(
            old_name="HistoricalMedicationStockCreateLabels",
            new_name="HistoricalStockCreateLabels",
        ),
        migrations.RenameModel(
            old_name="HistoricalMedicationStockReceiving",
            new_name="HistoricalStockReceiving",
        ),
        migrations.RenameModel(
            old_name="MedicationStock",
            new_name="Stock",
        ),
        migrations.RenameModel(
            old_name="MedicationStockCreateLabels",
            new_name="StockCreateLabels",
        ),
        migrations.RenameModel(
            old_name="MedicationStockReceiving",
            new_name="StockReceiving",
        ),
    ]



from django.core.management.base import BaseCommand

from edc_notification.site_notifications import site_notifications


class Command(BaseCommand):
    help = "List email recipients for each registered notification"

    def handle(self, *args, **options):
        for notification_cls in site_notifications.registry.values():
            notification = notification_cls()
            print()
            print(notification.name)
            print(notification.display_name)
            print(f"email: {notification.email_to}")



# from .aliquot_model_wrapper import AliquotModelWrapper
# from .box_model_wrapper import BoxModelWrapper
# from .manage_box_item_model_wrapper import ManageBoxItemModelWrapper
# from .manifest_item_model_wrapper import ManifestItemModelWrapper
# from .manifest_model_wrapper import ManifestModelWrapper
# from .requisition_model_wrapper import RequisitionModelWrapper
# from .result_model_wrapper import ResultModelWrapper
# from .verify_box_model_wrapper import VerifyBoxItemModelWrapper



from .aliquot_admin import AliquotAdmin
from .box_admin import BoxAdmin
from .box_item_admin import BoxItemAdmin
from .box_type_admin import BoxTypeAdmin
from .consignee_admin import ConsigneeAdmin
from .fieldsets import (
    requisition_fieldset,
    requisition_identifier_fields,
    requisition_identifier_fieldset,
    requisition_status_fields,
    requisition_status_fieldset,
    requisition_verify_fields,
    requisition_verify_fieldset,
)
from .manifest_admin import ManifestAdmin
from .manifest_item_admin import ManifestItemAdmin
from .modeladmin_mixins import RequisitionAdminMixin
from .order_admin import OrderAdmin
from .panel_admin import PanelAdmin
from .result_admin import ResultAdmin, ResultItemAdmin
from .shipper_admin import ShipperAdmin



# Generated by Django 3.2.13 on 2022-09-06 22:57

import uuid

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0011_auto_20220826_0406"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="rxrefill",
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name="historicalrxrefill",
            name="visit_code",
        ),
        migrations.RemoveField(
            model_name="historicalrxrefill",
            name="visit_code_sequence",
        ),
        migrations.RemoveField(
            model_name="rxrefill",
            name="visit_code",
        ),
        migrations.RemoveField(
            model_name="rxrefill",
            name="visit_code_sequence",
        ),
        migrations.RenameField(
            model_name="historicalrxrefill",
            old_name="refill_date",
            new_name="refill_start_datetime",
        ),
        migrations.RenameField(
            model_name="rxrefill",
            old_name="refill_date",
            new_name="refill_start_datetime",
        ),
    ]



from collections.abc import Iterable

from .row_definition import RowDefinition
from .row_statistics import RowStatistics
from .row_statistics_with_gender import RowStatisticsWithGender


class RowDefinitions:
    """Collection of RowDefinitions"""

    def __init__(
        self,
        colname: str = None,
        row_statistics_cls: RowStatistics | RowStatisticsWithGender = None,
        reverse_rows: bool = False,
    ):
        self.definitions: list[RowDefinition] = []
        self.row_statistics_cls = row_statistics_cls
        self.colname = colname
        self.reverse_rows = reverse_rows

    def add(self, row_definition: RowDefinition):
        self.definitions.append(row_definition)

    def extend(self, row_definition: list[RowDefinition]):
        self.definitions.extend(row_definition)

    def reverse(self):
        self.definitions.reverse()

    def __iter__(self) -> Iterable[RowDefinition]:
        return iter(self.definitions)



from __future__ import annotations

from dataclasses import dataclass, field


class AlreadyRegistered(Exception):  # noqa: N818
    pass


class InvalidDashboardUrlName(Exception):  # noqa: N818
    pass


@dataclass
class UrlNames:
    registry: dict[str, str] = field(default_factory=dict)

    def register(
        self, name: str | None = None, url: str | None = None, namespace: str | None = None
    ) -> None:
        name = name or url
        complete_url = f"{namespace}:{url}" if namespace else url
        if name in self.registry:
            raise AlreadyRegistered(f"Url already registered. Got {complete_url}.")
        self.registry.update({name: complete_url})

    def register_from_dict(self, **urldata: str) -> None:
        for name, complete_url in urldata.items():
            try:
                namespace, url = complete_url.split(":")
            except ValueError:
                namespace, url = complete_url, None
            self.register(name=name, url=url, namespace=namespace)

    def all(self) -> dict[str, str]:
        return self.registry

    def get(self, name: str) -> str:
        if name not in self.registry:
            raise InvalidDashboardUrlName(
                f"Invalid dashboard url name. Expected one of {self.registry.keys()}. "
                f"Got '{name}'."
            )
        return self.registry.get(name)

    def get_or_raise(self, name: str) -> str:
        return self.get(name)


url_names = UrlNames()



from __future__ import annotations

from django import forms

from edc_constants.constants import NO


class OgttFormValidatorMixin:
    def validate_ogtt_required_fields(
        self,
        ogtt_prefix: str | None = None,
        fasting_prefix: str | None = None,
    ) -> None:
        """Uses fields `fasting`, `ogtt_base_datetime`, `ogtt_datetime`,
        `ogtt_value`, `ogtt_units`
        """
        ogtt = ogtt_prefix or "ogtt"
        fasting = fasting_prefix or "fasting"
        self.required_if_true(
            self.cleaned_data.get(f"{ogtt}_datetime"),
            field_required=f"{ogtt}_value",
            inverse=False,
        )

        self.required_if_true(
            self.cleaned_data.get(f"{ogtt}_value"),
            field_required=f"{ogtt}_datetime",
            inverse=False,
        )

        self.not_required_if(
            NO,
            field=fasting,
            field_not_required=f"{ogtt}_base_datetime",
            inverse=False,
        )
        self.not_required_if(
            NO, field=fasting, field_not_required=f"{ogtt}_datetime", inverse=False
        )
        self.not_required_if(
            NO, field=fasting, field_not_required=f"{ogtt}_value", inverse=False
        )

        self.required_if_true(
            self.cleaned_data.get(f"{ogtt}_value"),
            field_required=f"{ogtt}_units",
        )

        self.not_required_if(
            NO, field=fasting, field_not_required=f"{ogtt}_units", inverse=False
        )

    def validate_ogtt_dates(self, ogtt_prefix: str | None = None) -> None:
        ogtt = ogtt_prefix or "ogtt"
        ogtt_base_dte = self.cleaned_data.get(f"{ogtt}_base_datetime")
        ogtt_dte = self.cleaned_data.get(f"{ogtt}_datetime")
        if ogtt_base_dte and ogtt_dte:
            tdelta = ogtt_dte - ogtt_base_dte
            if tdelta.total_seconds() < 3600:
                raise forms.ValidationError(
                    {
                        f"{ogtt}_datetime": (
                            "Invalid. Expected more time between OGTT initial and 2hr."
                        )
                    }
                )
            if tdelta.seconds > (3600 * 5):
                raise forms.ValidationError(
                    {
                        f"{ogtt}_datetime": (
                            "Invalid. Expected less time between OGTT initial and 2hr."
                        )
                    }
                )

    def validate_ogtt_time_interval(self, ogtt_prefix: str | None = None) -> None:
        """Validate the OGTT is measured 2 hrs after base date"""
        ogtt = ogtt_prefix or "ogtt"
        ogtt_base_dte = self.cleaned_data.get(f"{ogtt}_base_datetime")
        ogtt_dte = self.cleaned_data.get(f"{ogtt}_datetime")
        if ogtt_base_dte and ogtt_dte:
            diff = (ogtt_dte - ogtt_base_dte).total_seconds() / 60.0
            if diff <= 1.0:
                raise forms.ValidationError(
                    {
                        f"{ogtt}_datetime": (
                            "Invalid date. Expected to be after time oral glucose "
                            f"tolerance test was performed. ({diff})"
                        )
                    }
                )



from django.db import models
from django.db.models import Index


class NotificationMixin(models.Model):
    notification_plan_name = models.CharField(max_length=200)

    notification_datetime = models.DateTimeField()

    subject = models.CharField(max_length=200)

    recipient_list = models.TextField(null=True, default="")

    cc_list = models.TextField(null=True, default="")

    body = models.TextField(null=True, default="")

    status = models.CharField(
        max_length=15,
        default="new",
        choices=(("new", "New"), ("sent", "Sent"), ("cancelled", "Cancelled")),
    )

    sent = models.BooleanField(default=False)

    sent_datetime = models.DateTimeField(null=True)

    class Meta:
        abstract = True
        indexes = (Index(fields=["notification_datetime"]),)



from uuid import uuid4

from django.db import models
from django.utils import timezone

from edc_model.models import BaseUuidModel

from .room import Room


class Shelf(BaseUuidModel):
    shelf_identifier = models.CharField(max_length=36, default=uuid4, unique=True)

    shelf_datetime = models.DateTimeField(default=timezone.now)

    name = models.CharField(max_length=25, unique=True)

    description = models.TextField(default="")

    room = models.ForeignKey(Room, on_delete=models.PROTECT)

    def __str__(self):
        return f"Shelf {self.name}"

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Shelf"
        verbose_name_plural = "Shelves"



from django import forms

from ...models import Order, OrderItem


class OrderForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()
        if (
            getattr(self.instance, "id", None)
            and OrderItem.objects.filter(
                order=self.instance, receiveitem__isnull=False
            ).exists()
        ):
            raise forms.ValidationError(
                "Order cannot be changed. Some items have been received"
            )
        return cleaned_data

    class Meta:
        model = Order
        fields = "__all__"
        help_text = {"order_identifier": "(read-only)"}
        widgets = {
            "order_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }


class OrderFormSuper(forms.ModelForm):

    class Meta:
        model = Order
        fields = "__all__"
        help_text = {"order_identifier": "(read-only)"}
        widgets = {
            "order_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



# Generated by Django 4.2.7 on 2023-12-01 14:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_consent", "0002_alter_edcpermissions_device_created_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
    ]



from .allocate_stock import allocate_stock
from .blinded_user import blinded_user
from .confirm_stock import confirm_stock
from .confirm_stock_at_site import confirm_stock_at_site
from .create_new_stock_on_receive import create_new_stock_on_receive
from .dispense import dispense
from .format_qty import format_qty
from .get_codenames import get_codenames
from .get_imp_schedule_names import get_imp_schedule_names
from .get_random_code import get_random_code
from .get_related_or_none import get_related_or_none
from .get_stock_for_location_df import get_stock_for_location_df
from .get_unit_qty_out import get_unit_qty_out
from .miscellaneous import get_rx_model_cls, get_rxrefill_model_cls
from .process_repack_request import process_repack_request
from .process_repack_request_queryset import process_repack_request_queryset
from .stock_request import bulk_create_stock_request_items, get_instock_and_nostock_data
from .transfer_stock import transfer_stock
from .update_previous_refill_end_datetime import update_previous_refill_end_datetime
from .update_stock_instance import update_stock_instance

__all__ = [
    "allocate_stock",
    "blinded_user",
    "bulk_create_stock_request_items",
    "confirm_stock",
    "confirm_stock_at_site",
    "create_new_stock_on_receive",
    "dispense",
    "format_qty",
    "get_codenames",
    "get_imp_schedule_names",
    "get_instock_and_nostock_data",
    "get_random_code",
    "get_related_or_none",
    "get_rx_model_cls",
    "get_rxrefill_model_cls",
    "get_stock_for_location_df",
    "get_unit_qty_out",
    "process_repack_request",
    "process_repack_request_queryset",
    "transfer_stock",
    "update_previous_refill_end_datetime",
    "update_stock_instance",
]



import sys

from django.core.management.color import color_style

style = color_style()


def post_migrate_list_data(sender=None, **kwargs):
    from .site_list_data import get_autodiscover_enabled, site_list_data  # noqa: PLC0415

    if get_autodiscover_enabled():
        sys.stdout.write(style.MIGRATE_HEADING("Updating list data:\n"))
        site_list_data.load_data()
        sys.stdout.write("Done.\n")
        sys.stdout.flush()



from django.db import models

from edc_constants.choices import YES_NO_NA
from edc_constants.constants import NOT_APPLICABLE

from .followup_review_model_mixin import FollowupReviewModelMixin


class HivFollowupReviewModelMixin(FollowupReviewModelMixin, models.Model):
    dx = models.CharField(
        verbose_name="Has the patient been infected with HIV?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    arv_initiated = models.CharField(
        verbose_name="Has the patient started antiretroviral therapy (ART)?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
        help_text="Select `not applicable` if previously reported.",
    )
    arv_initiation_actual_date = models.DateField(
        verbose_name="Date started antiretroviral therapy (ART)",
        null=True,
        blank=True,
    )

    class Meta(FollowupReviewModelMixin.Meta):
        abstract = True
        verbose_name = "HIV Review"
        verbose_name_plural = "HIV Review"



from edc_form_runners.models import Issue as BaseIssue


class Issue(BaseIssue):
    class Meta:
        proxy = True
        verbose_name = "Form validation issue"
        verbose_name_plural = "Form validation issues"



CANCELLED = "cancelled"
CSV = "CSV"
DELETE = "D"
EXPORTED = "exported"
INSERT = "I"
UPDATE = "U"
STATA_10 = 114
STATA_13 = 117
STATA_14 = 118
STATA_15 = 119

EXPORT = "EXPORT"
EXPORT_PII = "EXPORT_PII"
DATA_EXPORTER_ROLE = "data_exporter"



# Generated by Django 5.1 on 2024-08-14 17:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0012_alter_qareportnote_note"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="QaReportNote",
            new_name="Note",
        ),
        migrations.RemoveConstraint(
            model_name="note",
            name="edc_qareports_qareportnote_report_model_subj_uniq",
        ),
        migrations.RenameIndex(
            model_name="note",
            new_name="edc_qarepor_modifie_90fe6c_idx",
            old_name="edc_qarepor_modifie_5f102f_idx",
        ),
        migrations.RenameIndex(
            model_name="note",
            new_name="edc_qarepor_user_mo_d0d80a_idx",
            old_name="edc_qarepor_user_mo_6166bf_idx",
        ),
        migrations.AddConstraint(
            model_name="note",
            constraint=models.UniqueConstraint(
                fields=("report_model", "subject_identifier"),
                name="edc_qareports_note_report_model_subj_uniq",
            ),
        ),
    ]



# Generated by Django 5.2.6 on 2025-09-17 16:53

import edc_model_fields.fields.other_charfield
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_visit_tracking", "0008_alter_historicalsubjectvisit_revision_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="comments",
            field=models.TextField(
                blank=True,
                default="",
                max_length=250,
                verbose_name="Comment if any additional pertinent information about the participant",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="consent_model",
            field=models.CharField(blank=True, default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="consent_version",
            field=models.CharField(blank=True, default="", max_length=10),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="document_status_comments",
            field=models.TextField(
                blank=True,
                default="",
                help_text="for example, why some data is still pending",
                verbose_name="Any comments related to status of this document",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="info_source_other",
            field=edc_model_fields.fields.other_charfield.OtherCharField(
                blank=True,
                default="",
                max_length=35,
                verbose_name='If "Other" source of information, specify',
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="reason_missed",
            field=models.CharField(
                blank=True,
                choices=[
                    ("timepoint", "Missed timepoint"),
                    ("OTHER", "Other"),
                    ("N/A", "Not applicable"),
                ],
                default="",
                max_length=35,
                verbose_name="If 'missed', provide the reason for the missed visit",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="reason_missed_other",
            field=edc_model_fields.fields.other_charfield.OtherCharField(
                blank=True,
                default="",
                max_length=25,
                verbose_name='If the reason for the missed visit is "other", specify',
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="study_status",
            field=models.CharField(
                default="",
                max_length=50,
                verbose_name="What is the participant's current study status",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="survival_status",
            field=models.CharField(
                choices=[
                    ("alive", "Alive"),
                    ("dead", "Deceased"),
                    ("unknown", "Unknown"),
                    ("N/A", "Not applicable (if missed)"),
                ],
                default="alive",
                help_text="If YES, submit Death report",
                max_length=10,
                verbose_name="Participant's survival status",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisit",
            name="visit_code",
            field=models.CharField(default="", editable=False, max_length=25),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="comment",
            field=models.TextField(
                blank=True,
                default="",
                verbose_name="Please provide further details, if any",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="consent_model",
            field=models.CharField(blank=True, default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="consent_version",
            field=models.CharField(blank=True, default="", max_length=10),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="contact_attempts_explained",
            field=models.TextField(
                blank=True,
                default="",
                verbose_name="If contact not made and less than 3 attempts, please explain",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectvisitmissed",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="comments",
            field=models.TextField(
                blank=True,
                default="",
                max_length=250,
                verbose_name="Comment if any additional pertinent information about the participant",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="consent_model",
            field=models.CharField(blank=True, default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="consent_version",
            field=models.CharField(blank=True, default="", max_length=10),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="document_status_comments",
            field=models.TextField(
                blank=True,
                default="",
                help_text="for example, why some data is still pending",
                verbose_name="Any comments related to status of this document",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="info_source_other",
            field=edc_model_fields.fields.other_charfield.OtherCharField(
                blank=True,
                default="",
                max_length=35,
                verbose_name='If "Other" source of information, specify',
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="reason_missed",
            field=models.CharField(
                blank=True,
                choices=[
                    ("timepoint", "Missed timepoint"),
                    ("OTHER", "Other"),
                    ("N/A", "Not applicable"),
                ],
                default="",
                max_length=35,
                verbose_name="If 'missed', provide the reason for the missed visit",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="reason_missed_other",
            field=edc_model_fields.fields.other_charfield.OtherCharField(
                blank=True,
                default="",
                max_length=25,
                verbose_name='If the reason for the missed visit is "other", specify',
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="study_status",
            field=models.CharField(
                default="",
                max_length=50,
                verbose_name="What is the participant's current study status",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="survival_status",
            field=models.CharField(
                choices=[
                    ("alive", "Alive"),
                    ("dead", "Deceased"),
                    ("unknown", "Unknown"),
                    ("N/A", "Not applicable (if missed)"),
                ],
                default="alive",
                help_text="If YES, submit Death report",
                max_length=10,
                verbose_name="Participant's survival status",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisit",
            name="visit_code",
            field=models.CharField(default="", editable=False, max_length=25),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="comment",
            field=models.TextField(
                blank=True,
                default="",
                verbose_name="Please provide further details, if any",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="consent_model",
            field=models.CharField(blank=True, default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="consent_version",
            field=models.CharField(blank=True, default="", max_length=10),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="contact_attempts_explained",
            field=models.TextField(
                blank=True,
                default="",
                verbose_name="If contact not made and less than 3 attempts, please explain",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissed",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissedreasons",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissedreasons",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissedreasons",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
        migrations.AlterField(
            model_name="visitreasons",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="visitreasons",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="visitreasons",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
    ]



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ...admin_site import edc_pharmacy_admin
from ...forms import DosageGuidelineForm
from ...models import DosageGuideline
from ..model_admin_mixin import ModelAdminMixin


@admin.register(DosageGuideline, site=edc_pharmacy_admin)
class DosageGuidelineAdmin(ModelAdminMixin, admin.ModelAdmin):
    show_object_tools = True

    autocomplete_fields = ["medication"]

    form = DosageGuidelineForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "medication",
                    "dose",
                    "dose_per_kg",
                    "dose_units",
                    "frequency",
                    "frequency_units",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display: tuple[str, ...] = ("__str__", "modified", "user_modified")
    search_fields: tuple[str, ...] = ("medication__name",)



from django.contrib.admin import SimpleListFilter
from django.db.models import Count, QuerySet

from edc_constants.constants import NEW

from ..choices import NOTE_STATUSES


class NoteStatusListFilter(SimpleListFilter):
    title = "QA Status"
    parameter_name = "note_status"

    note_model_cls = None
    note_model_status_choices = NOTE_STATUSES

    def __init__(self, request, params, model, model_admin):
        self.note_model_cls = model_admin.note_model_cls
        super().__init__(request, params, model, model_admin)

    def lookups(self, request, model_admin):
        status_dict = {tpl[0]: tpl[1] for tpl in self.note_model_status_choices}
        names = [(NEW, status_dict.get(NEW, "New"))]
        qs = (
            self.note_model_cls.objects.values("status")
            .order_by("status")
            .annotate(cnt=Count("status"))
        )

        for obj in qs:
            names.append((f"{obj.get('status')}", status_dict[obj.get("status")]))
        return tuple(names)

    @staticmethod
    def report_model(queryset: QuerySet) -> str:
        qs = (
            queryset.values("report_model")
            .order_by("report_model")
            .annotate(cnt=Count("report_model"))
        )
        for obj in qs:
            return obj.get("report_model")

    def queryset(self, request, queryset):
        if self.value() and self.value() != "none":
            if report_model := self.report_model(queryset):
                if self.value() == NEW:
                    qs = self.note_model_cls.objects.values("subject_identifier").filter(
                        report_model=report_model
                    )
                    queryset = queryset.exclude(
                        subject_identifier__in=[obj.get("subject_identifier") for obj in qs]
                    )
                elif self.value() in [tpl[0] for tpl in self.note_model_status_choices]:
                    qs = self.note_model_cls.objects.values("subject_identifier").filter(
                        report_model=report_model, status=self.value()
                    )
                    queryset = queryset.filter(
                        subject_identifier__in=[obj.get("subject_identifier") for obj in qs]
                    )
        return queryset



# Generated by Django 2.0.7 on 2018-07-06 13:19

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_appointment", "0014_auto_20180116_1411")]

    operations = [
        migrations.AlterModelOptions(
            name="historicalappointment",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical appointment",
            },
        )
    ]



import warnings

from .utils import get_ae_model, get_ae_model_name  # noqa

warnings.warn(
    "This path is deprecated in favor of edc_adverse_event.utils.",
    DeprecationWarning,
    stacklevel=2,
)



# Generated by Django 4.2.3 on 2023-08-16 15:25

from django.db import migrations

import edc_model_fields.fields.other_charfield


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0039_appointmenttype_extra_value"),
    ]

    operations = [
        migrations.AddField(
            model_name="appointment",
            name="appt_type_other",
            field=edc_model_fields.fields.other_charfield.OtherCharField(
                blank=True,
                max_length=35,
                null=True,
                verbose_name="If other appointment type, please specify ...",
            ),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="appt_type_other",
            field=edc_model_fields.fields.other_charfield.OtherCharField(
                blank=True,
                max_length=35,
                null=True,
                verbose_name="If other appointment type, please specify ...",
            ),
        ),
    ]



from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin

from ...exceptions import StockTransferConfirmationError
from .location import Location
from .stock_transfer import StockTransfer


class Manager(models.Manager):
    use_in_migrations = True


class ConfirmationAtSite(SiteModelMixin, BaseUuidModel):
    transfer_confirmation_identifier = models.CharField(
        max_length=36,
        unique=True,
        blank=True,
        null=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    transfer_confirmation_datetime = models.DateTimeField(default=timezone.now)

    stock_transfer = models.OneToOneField(StockTransfer, on_delete=models.PROTECT)

    location = models.ForeignKey(
        Location, on_delete=models.PROTECT, limit_choices_to={"site__isnull": False}
    )

    comments = models.TextField(default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.transfer_confirmation_identifier

    def save(self, *args, **kwargs):
        if self.location != self.stock_transfer.to_location:
            raise StockTransferConfirmationError(
                "Location mismatch. Perhaps catch this in the form."
            )
        self.site = self.location.site
        if not self.transfer_confirmation_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.transfer_confirmation_identifier = f"{next_id:06d}"
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock Confirmation at Site "
        verbose_name_plural = "Stock Confirmations at Site"



import pandas as pd

from ..table_to_dataframe import Helper, TableToDataframe


class Table:
    default_columns = None
    helper_cls = Helper
    table_to_dataframe_cls = TableToDataframe

    def __init__(
        self,
        table_name=None,
        columns=None,
        filename=None,
        local_tz=None,
        limit=None,
        **kwargs,
    ):
        self.dataframe = None
        self.helper = self.helper_cls(local_tz=local_tz)
        self.columns = columns or self.default_columns
        if filename:
            self.read_csv(filename=filename)
        else:
            self.dataframe = self.table_to_dataframe_cls(
                table_name=table_name, columns=self.columns, limit=limit
            ).dataframe
            self.prepare(**kwargs)

    def prepare(self, **kwargs):
        pass

    def read_csv(self, filename=None):
        self.dataframe = pd.read_csv(filename, low_memory=False)



from __future__ import annotations

from uuid import uuid4

from django.conf import settings

from ..utils import get_duplicates
from .crf import Crf
from .requisition import Requisition


class FormsCollectionError(Exception):
    pass


class FormsCollection:
    def __init__(
        self,
        *forms: Crf | Requisition,
        name: str | None = None,
        check_sequence: bool | None = None,
        **kwargs,  # noqa: ARG002
    ):
        check_sequence = True if check_sequence is None else check_sequence
        self.collection_is_unique_or_raise(forms)
        self._forms: tuple[Crf | Requisition] | None = None
        self.name = name or uuid4().hex
        forms = [] if not forms or forms == (None,) else list(forms)

        # exclude any flagged for a site that is not the current
        forms = [f for f in forms if not f.site_ids or settings.SITE_ID in f.site_ids]

        # sort on show order
        try:
            forms.sort(key=lambda x: x.show_order)
        except AttributeError as e:
            raise FormsCollectionError(e) from e

        # check sequence
        if check_sequence:
            seq = [item.show_order for item in forms or []]
            if duplicates := get_duplicates(list_items=seq):
                raise FormsCollectionError(
                    f'{self.__class__.__name__} "show order" must be a '
                    f"unique sequence. Got {sorted(seq)}. "
                    f"Duplicates {sorted(duplicates)}. See {self}."
                )

        # convert to tuple
        self._forms = tuple(forms)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"

    def __iter__(self):
        return iter(self._forms)

    def __len__(self):
        return len(self._forms)

    @staticmethod
    def collection_is_unique_or_raise(forms):
        pass

    def append(self, value):
        if value:
            forms = list(self._forms)
            for item in forms:
                if item.name == value.name:
                    raise FormsCollectionError(
                        f"Append failed. Item is not unique. Got {value.name}"
                    )
            forms.append(value)
            forms.sort(key=lambda x: x.show_order)
            self._forms = forms

    def extend(self, value: tuple | list):
        if value:
            for v in value:
                self.append(v)
            forms = list(self._forms)
            forms.sort(key=lambda x: x.show_order)
            self._forms = tuple(forms)

    def insert(self, index, value):
        if value:
            forms = list(self._forms)
            for _, item in forms:
                if item.name == value.name:
                    raise FormsCollectionError(
                        f"Insert failed. Item is not unique. Got {value.name}"
                    )
            forms.insert(index, value)
            forms.sort(key=lambda x: x.show_order)
            self._forms = tuple(forms)

    def remove(self, value):
        if value:
            forms = list(self._forms)
            for index, item in enumerate(forms):
                if item.name == value.name:
                    forms.pop(index)
                    self._forms = tuple(forms)
                    break
            else:
                raise FormsCollectionError("Remove failed. Item not found")

    def pop(self, index):
        forms = list(self._forms)
        forms.pop(index)
        self._forms = tuple(forms)

    def insert_last(self, value):
        forms = list(self._forms)
        value.show_order = 100 + max([item.show_order for item in forms or []])
        self.append(value)

    @property
    def forms(self) -> tuple:
        return self._forms



# Generated by Django 3.0.4 on 2020-05-12 21:34

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0020_auto_20191024_1000"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="aliquot",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Aliquot",
            },
        ),
        migrations.AlterModelOptions(
            name="box",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-box_datetime",),
                "verbose_name": "Box",
                "verbose_name_plural": "Boxes",
            },
        ),
        migrations.AlterModelOptions(
            name="boxitem",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("position",),
                "verbose_name": "Box Item",
            },
        ),
        migrations.AlterModelOptions(
            name="boxtype",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("name",),
                "verbose_name": "Box Type",
            },
        ),
        migrations.AlterModelOptions(
            name="consignee",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("name",),
                "verbose_name": "Consignee",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalaliquot",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Aliquot",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalbox",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Box",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalboxitem",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Box Item",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconsignee",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Consignee",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmanifest",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Manifest",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalorder",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Order",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalresult",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Result",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalresultitem",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Result Item",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalshipper",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Shipper",
            },
        ),
        migrations.AlterModelOptions(
            name="manifest",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-manifest_identifier",),
                "verbose_name": "Manifest",
            },
        ),
        migrations.AlterModelOptions(
            name="manifestitem",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("created",),
                "verbose_name": "Manifest Item",
            },
        ),
        migrations.AlterModelOptions(
            name="order",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Order",
            },
        ),
        migrations.AlterModelOptions(
            name="panel",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("lab_profile_name", "name"),
                "verbose_name": "Panel",
            },
        ),
        migrations.AlterModelOptions(
            name="result",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Result",
            },
        ),
        migrations.AlterModelOptions(
            name="resultitem",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Result Item",
            },
        ),
        migrations.AlterModelOptions(
            name="shipper",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("name",),
                "verbose_name": "Shipper",
            },
        ),
    ]
